<?php
/**
 * Filemanager - history and notifications
 *
 * @link http://www.egroupware.org
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @package tracker
 * @copyright (c) 2011 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id: class.filemanager_tracking.inc.php 2503 2013-03-20 09:54:26Z ralfbecker $
 */

/**
 * Filemanager - tracking object for filemanager
 */
class filemanager_tracking extends bo_tracking
{
	/**
	 * Application we are tracking (required!)
	 *
	 * @var string
	 */
	var $app = 'filemanager';
	/**
	 * Name of the id-field, used as id in the history log (required!)
	 *
	 * @var string
	 */
	var $id_field;// = 'fs_id';
	/**
	 * Name of the field with the creator id, if the creator of an entry should be notified
	 *
	 * @var string
	 */
	var $creator_field;
	/**
	 * Name of the field with the id(s) of assinged users, if they should be notified
	 *
	 * @var string
	 */
	var $assigned_field = 'subscribers';
	/**
	 * Translate field-name to 2-char history status
	 *
	 * @var array
	 */
	var $field2history = array();
	/**
	 * Should the user (passed to the track method or current user if not passed) be used as sender or get_config('sender')
	 *
	 * @var boolean
	 */
	var $prefer_user_as_sender = true;

	/**
         * Should the current user be email-notified (about change he made himself)
         *
         * @var boolean
         */
        var $notify_current_user = true;

	/**
	 * Constructor
	 *
	 * @param tracker_bo $botracker
	 * @return tracker_tracking
	 */
	function __construct()
	{
		parent::__construct('filemanager');	// adding custom fields for tracker

		// setting curent user, to be used as sender
		$this->user = $GLOBALS['egw_info']['user']['account_id'];
	}

	/**
	 * Get the subject for a given entry, can be reimplemented
	 *
	 * @param array $data
	 * @param array $old
	 * @param boolean $deleted=null can be set to true to let the tracking know the item got deleted or undelted
	 * @param int|string $receiver nummeric account_id or email address
	 * @return string
	 */
	protected function get_subject($data,$old,$deleted=null,$receiver=null)
	{
		if (empty($data['event']) && $deleted) $data['event'] = 'vfs_unlink';

		// some chars are internally urlencoded, decode them for the user
		$path = urldecode($data['path']);

		switch($data['event'])
		{
			case 'vfs_added':
				return lang('File %1 added',$path);
			case 'vfs_unlink':
				return lang('File %1 deleted',$path);
			case 'vfs_modified':
			default:
				return lang('File %1 modified',$path);
		}
	}

	/**
	 * Get the details of an entry
	 *
	 * @param array $data stat array
	 * @param int|string $receiver nummeric account_id or email address
	 * @return array of details as array with values for keys 'label','value','type'
	 */
	function get_details($data,$receiver=null)
	{
		static $labels = array(
			'size' => 'Size',
			'mime' => 'Type',
			'mtime'=> 'Modified',
			'ctime'=> 'Created',
			'uid'  => 'Owner',
			'gid'  => 'Group',
			'mode' => 'Permissions',
			'comment' => 'Comment',
		);
		if (!isset($data['mime'])) $data['mime'] = egw_vfs::mime_content_type($data['path']);

		// get comment and custom fields, if not already supplied
		if (!array_key_exists('comment', $data) && ($props = egw_vfs::propfind($data['path'],egw_vfs::DEFAULT_PROP_NAMESPACE)))
		{
			foreach($props as $prop)
			{
				$data[$prop['name']] = $prop['val'];
			}
		}

		$details = array();
		foreach(array(
			'mtime'=> egw_time::to($data['mtime']),
			'comment' => $data['comment'],
			'cfs'  => true,
			'ctime'=> egw_time::to($data['ctime']),
			'size' => egw_vfs::hsize($data['size']),
			'mime' => mime_magic::mime2label($data['mime']),
			'uid'  => $data['uid'] ? common::grab_owner_name($data['uid']) : 'root',
			'gid'  => $data['gid'] ? common::grab_owner_name(-$data['gid']) : 'root',
			'mode' => egw_vfs::int2mode($data['mode']),
		) as $name => $value)
		{
			if ($name == 'cfs')
			{
				$details += $this->get_customfields($data);
				continue;
			}
			$details[$name] = array(
				'label' => lang($labels[$name]),
				'value' => $value,
			);
		}
		//error_log(__METHOD__.'('.array2string($data).') returning '.array2string($details));
		return $details;
	}

	/**
	 * Get the body of the notification message, can be reimplemented
	 *
	 * Reimplemented to include the link with label "Download" for added or modified files notification to users, none otherwise
	 *
	 * @param boolean $html_email
	 * @param array $data
	 * @param array $old
	 * @param boolean $integrate_link to have links embedded inside the body
	 * @param int|string $receiver nummeric account_id or email address
	 * @return string
	 */
	public function get_body($html_email,$data,$old,$integrate_link = true,$receiver=null)
	{
		if (!in_array($data['event'],array('vfs_added','vfs_modified')) || strpos($receiver,'@') !== false)
		{
			$integrate_link = false;
		}
		else
		{
			$integrate_link = lang('Download');
		}
		return parent::get_body($html_email, $data, $old, $integrate_link, $receiver);
	}

	/**
	 * Get the modified / new message (1. line of mail body) for a given entry, can be reimplemented
	 *
	 * Reimplemented to add message "Added/modified file is attached" for external receipents and vfs_added|modfied
	 *
	 * @param array $data
	 * @param array $old
	 * @param int|string $receiver nummeric account_id or email address
	 * @return string
	 */
	protected function get_message($data,$old,$receiver=null)
	{
		$msg = '';

		if (strpos($receiver,'@') !== false)
		{
			switch($data['event'])
			{
				case 'vfs_added':
					$msg = lang('Added file is attached to this mail.');
					break;
				case 'vfs_modified':
					$msg = lang('Modified file is attached to this mail.');
					break;
			}
		}
		return $msg;
	}

	/**
	 * Get a config value, which can depend on $data and $old
	 *
	 * @param string $what possible values are:
	 * 	- 'copy' array of email addresses notifications should be copied too, can depend on $data
	 *  - 'lang' string lang code for copy mail
	 *  - 'subject' string subject line for the notification of $data,$old, defaults to link-title
	 *  - 'link' string of link to view $data
	 *  - 'sender' sender of email
	 * @param array $data current entry
	 * @param array $old=null old/last state of the entry or null for a new entry
	 * @return mixed
	 */
	protected function get_config($name,$data,$old=null)
	{
		switch($name)
		{
			case 'link':
				// no download link for deleted files!
				if (in_array($data['event'], array('vfs_modified','vfs_added')))
				{
					$ret = egw::link(egw_vfs::download_url($data['path'], true));
				}
				break;
			case 'link_no_id':
				// do NOT add id to link, as it breaks WebDAV downloads in certain cases, beside being unnecessary
				$ret = true;
				break;
			case 'copy':
				$ret = $data['email_subs'];
				break;
		}
		//error_log(__METHOD__."('$name', ".array2string($data).', '.array2string($old).') returning '.array2string($ret));
		return $ret;
	}

	/**
	 * Sending a notification to the given email-address
	 *
	 * Reimplemented to set $check to null, as we use a subscription model for the notifications and not a preferences.
	 *
	 * @param array $data current entry
	 * @param array $old=null old/last state of the entry or null for a new entry
	 * @param string $email address to send the notification to
	 * @param string $user_or_lang='en' user-id or 2 char lang-code for a non-system user
	 * @param string $check=null pref. to check if a notification is wanted
	 * @param boolean $assignment_changed=true the assignment of the user $user_or_lang changed
	 * @param boolean $deleted=null can be set to true to let the tracking know the item got deleted or undelted
	 * @return boolean true on success or false on error (error-message is in $this->errors)
	 */
	public function send_notification($data,$old,$email,$user_or_lang,$check=null,$assignment_changed=true,$deleted=null)
	{
		return parent::send_notification($data, $old, $email, $user_or_lang, null, $assignment_changed, $deleted);
	}

	/**
	 * Get the attachments for a notification
	 *
	 * @param array $data
	 * @param array $old
	 * @param int|string $receiver nummeric account_id or email address
	 * @return array or array with values for either 'string' or 'path' and optionally (mime-)'type', 'filename' and 'encoding'
	 */
	protected function get_attachments($data,$old,$receiver=null)
	{
		$attachments = array();
		// add file as attachment only for non-users
		if (strpos($receiver,'@') !== false && in_array($data['event'], array('vfs_modified','vfs_added')))
		{
			$attachments[] = array(
				'path' => egw_vfs::PREFIX.$data['path'],
				'type' => $data['mime'],
				'filename' => egw_vfs::basename($data['path']),
				//'string' => file_get_contents(egw_vfs::PREFIX.$data['path']),
			);
		}
		//error_log(__METHOD__.'('.array2string($data).",,'$receivers') returning ".array2string($attachments));
		return $attachments;
	}
}
