<?php
/**
 * Stylite: Filemanager - favorites
 *
 * @link http://www.stylite.de
 * @package stylite
 * @subpackage filemanager
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2009-15 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id: class.stylite_filemanager_favorites.inc.php 3666 2015-04-07 10:01:31Z ralfbecker $
 */

use EGroupware\Stylite\Vfs\Links;

/**
 * Favorites for filemanager
 *
 * The favorites get stored as symlinks in two hidden directories:
 * - /home/.favorites      for all users displayed on top
 * - /home/$user/.favorite for user $user (only these can be created/deleted via the GUI)
 */
class stylite_filemanager_favorites implements Links\PluginIface
{
	static $appname = 'filemanager';
	static $favorites = array();
	/**
	 * Dir to store user specific favorites
	 *
	 * @var string
	 */
	static $user_fav_dir;
	const SYSTEM_FAV_DIR = '/home/.favorites';

	/**
	 * Data for Filemanagers sidebox menu
	 *
	 * @param array $args
	 */
	public static function sidebox_filemanager($args)
	{
		unset($GLOBALS['egw_info']['user']['preferences']['common']['auto_hide_sidebox']);

		// Move old favorites to new
		$pref_prefix = 'favorite_';
		foreach(self::$favorites as $n => $path)
		{
			if (is_array($path))
			{
				$name = $path['name'];
				$app  = $path['app'];
				$icon = $path['icon'];
				$title = $path['title'];
				$no_delete = $path['no_delete'];
				$path = $path['path'];
			}
			else
			{
				$name = egw_vfs::basename($path);
				$app  = 'etemplate';
				$icon = 'mime16_directory';
				$title = $path;
				$no_delete = false;
			}
			$safe_name = preg_replace('/[^A-Za-z0-9-_]/',"_",$title);
			$GLOBALS['egw']->preferences->add(self::$appname,$pref_prefix . $safe_name, array(
				'state' => array('path' => $path,'filter' => 1),
				'name' => $name,
				'group' => ''
			));

		}
		$GLOBALS['egw']->preferences->save_repository();
		return true;
	}

	/**
	 * Actions called via /stylite/favorites.php?app=filemanager with currentapp='filemanager'
	 *
	 */
	public static function action()
	{
		switch($method=$_GET['action'])
		{
			case 'add':
			case 'delete':
				self::$method();
				break;
			case 'sidebox_filemanager':
			default:
				throw new egw_exception_wrong_parameter(__METHOD__."() Unknown action=".$_GET['action']);
				break;
		}
	}

	/**
	 * Delete favorite specified by _GET[id]
	 */
	private static function delete()
	{
		$favorite = self::$favorites[$_GET['id']];
		error_log(__METHOD__." -> delete:".$favorite);
		if (!isset($favorite))
		{
			throw new egw_exception_assertion_failed('Unknown id to delete!');
		}
		unset(self::$favorites[$_GET['id']]);
		self::$favorites = array_values(self::$favorites);	// reindex
		egw_cache::setSession(__CLASS__,'favorites',self::$favorites);

		if (is_array($favorite)) $favorite = $favorite['name'];
		$favorite = egw_vfs::basename($favorite);
		//error_log(self::$user_fav_dir.'/'.$favorite);
		if ($favorite && egw_vfs::unlink(self::$user_fav_dir.'/'.$favorite))
		{
			$msg = lang('Favorite deleted.');
		}
		else
		{
			$msg = lang('An error happend');
		}
		egw::redirect_link('/index.php',array(
			'menuaction' => 'filemanager.filemanager_ui.index',
			'msg' => $msg,
		));
	}

	/**
	 * Add currently displayed path as favorite
	 *
	 * @param string $path=null path to add or default NULL to use path from session
	 * @param boolean $no_delete=false
	 * @param string $name=null
	 */
	private static function add($path=null,$no_delete=false,$name=null)
	{
		if (is_null($path))
		{
			$query = egw_session::appsession('index','filemanager');
			if ($query === false || !isset($query['path']) || !egw_vfs::is_readable($path=$query['path']))
			{
				throw new egw_exception_assertion_failed('Currently no (valid) path set by filemanager');
			}
		}
		if (!in_array($path,self::$favorites))
		{
			static $places = array(
				'/' => 'Basedirectory',
				'/home' => 'Users and groups',
				'/apps' => 'Applications',
			);
			list(,$apps,$app,$rest) = explode('/',$path);
			if(isset($places[$path]))
			{
				$path = array(
					'name' => lang($places[$path]),
					'path' => $path,
					'icon' => 'navbar',
					'app'  => 'filemanager',
					'title'=> $path,
				);
			}
			elseif ($apps == 'apps' && is_null($rest))	// normalize application directory
			{
				if (!isset($GLOBALS['egw_info']['apps'][$app]))
				{
					foreach($GLOBALS['egw_info']['apps'] as $name => $data)
					{
						if ($data['title'] == $app)
						{
							$app = $name;
							break;
						}
					}
				}
				if (isset($GLOBALS['egw_info']['apps'][$app]))
				{
					$path = array(
						'name' => $GLOBALS['egw_info']['apps'][$app]['title'],
						'path' => '/apps/'.$app,
						'icon' => 'navbar',
						'app'  => $app,
						'title'=> '/'.lang('Applications').'/'.$GLOBALS['egw_info']['apps'][$app]['title'],
					);
				}
			}
			elseif($apps == 'home' && is_null($rest))
			{
				if ($app == $GLOBALS['egw_info']['user']['account_lid'])
				{
					$path = array(
						'name' => lang('Your home directory'),
						'path' => $path,
						'icon' => 'navbar',
						'app'  => 'home',
						'title'=> $path,
					);
				}
				else
				{
					$path = array(
						'name' => ($GLOBALS['egw']->accounts->name2id($app) > 0 ? lang('User') : lang('Group')).' '.$app,
						'path' => $path,
						'icon' => 'users',
						'title'=> $path,
					);
				}
			}
			if ($no_delete || !is_array($path))
			{
				if (!is_array($path))
				{
					list($app,$icon) = explode('/',egw_vfs::mime_icon(egw_vfs::mime_content_type($path)));
					$path = array(
						'name' => isset($name) ? $name : egw_vfs::basename($path),
						'path' => $path,
						'title' => $path,
						'app'  => $app,
						'icon' => $icon,
					);
				}
				$path['no_delete'] = $no_delete;
			}

			if (isset($query))	// add called by user?
			{
				$rv = self::save(is_array($path) ? $path['path'] : $path,is_array($path) ? $path['name'] : null);
				//error_log($rv);
				if ($rv)
				{
					$msg = lang('Favorite added.')." ".$rv;
					$path['name'] = $rv;
				}
				else
				{
					$msg = lang('An error happend');
				}
			}
			self::$favorites[] = $path;
			if (isset($query)) egw_cache::setSession(__CLASS__,'favorites',self::$favorites);
		}
		if (isset($query))
		{
			egw::redirect_link('/index.php',array(
				'menuaction' => 'filemanager.filemanager_ui.index',
				'msg' => $msg,
			));
		}
	}

	/**
	 * Save a (user specific) favorite
	 *
	 * @param string $target
	 * @param string $name=null default use basename($target)
	 * @return 0 or string (name of the link)
	 */
	private static function save($target,$name=null)
	{
		//error_log(__METHOD__." favorite ".print_r($target,true)." saving with name:".$name);
		if (!egw_vfs::file_exists(self::$user_fav_dir)) egw_vfs::mkdir(self::$user_fav_dir,070,0);

		if(is_null($name)) $name = egw_vfs::basename($target);
		if (egw_vfs::file_exists(self::$user_fav_dir.'/'.$name))
		{
			for ($i=1; $i<99; $i++)
			{
				$namemod = $name."($i)";
				if (!egw_vfs::file_exists(self::$user_fav_dir.'/'.$namemod))
				{
					egw_vfs::symlink($target,self::$user_fav_dir.'/'.$namemod);
					return $namemod;
				}
			}
			error_log(self::$user_fav_dir.'/'.$name." already exists. No action taken.");
			return 0;
		}
		egw_vfs::symlink($target,self::$user_fav_dir.'/'.$name);
		return $name;
	}

	/**
	 * Initialisie our static variables
	 *
	 */
	public static function init_static()
	{
		self::$user_fav_dir = '/home/'.$GLOBALS['egw_info']['user']['account_lid'].'/.favorites';

		if (!(self::$favorites = egw_cache::getSession(__CLASS__,'favorites')))
		{
			self::$favorites = array();
			foreach(array(self::SYSTEM_FAV_DIR,self::$user_fav_dir) as $base)
			{
				if (egw_vfs::file_exists($base) && ($dir = egw_vfs::opendir($base)))
				{
					while(($name = readdir($dir)))
					{
						$path = egw_vfs::readlink($base.'/'.$name);
						self::add($path,$base == self::SYSTEM_FAV_DIR,$name);
					}
					closedir($dir);
				}
			}
			// import old favorites, can ge removed soon!
			if (!self::$favorites && isset($GLOBALS['egw_info']['user']['preferences']['filemanager']['favorites']))
			{
				self::$favorites = $GLOBALS['egw_info']['user']['preferences']['filemanager']['favorites'];
				foreach(self::$favorites as $path)
				{
					if (is_array($path))
					{
						$name = $path['name'];
						$path = $path['path'];
					}
					else
					{
						$name = egw_vfs::basename($path);
					}
					self::save($path,$name);
				}
			}
			egw_cache::setSession(__CLASS__,'favorites',self::$favorites);
		}
	}

	/**
	 * Stat something in the application directory
	 *
	 * An implementation can eg. call the appdir method of one level higher, to populate the stat cache
	 * with everything allowed on that level --> Links\StreamWrapper::url_stat then looks up the cache
	 *
	 * With late static binding (php5.3+) this could be done by adding the code here and
	 * converting the interface into an abstract base class.
	 *
	 * @param string $id filename in the application directory
	 */
	static public function stat($id)
	{
		self::appdir('');
		//echo "<p>".__METHOD__."($id)=".array2string(Links\StreamWrapper::getStatCache('favorites/'.$id))."</p>\n";
	}

	/**
	 * Return directory entries for application and populate the stat cache
	 *
	 * @param string $id filename in the application directory
	 * @return array with entries to return to the user
	 */
	static public function appdir($id)
	{
		//echo "<p>".__METHOD__."($id)</p>\n";
		if ($id != '') return null;

		$entries = array();
		foreach(self::$favorites as $favorite)
		{
			if (!is_array($favorite))
			{
				$favorite = array(
					'name' => egw_vfs::basename($favorite),
					'path' => $favorite,
				);
			}
			$entries[] = $favorite['name'];
			Links\StreamWrapper::addStatCacheLink('favorites/'.$favorite['name'],$favorite['path']);
		}
		return $entries;
	}
}

stylite_filemanager_favorites::init_static();
