<?php
/**
 * Stylite: VFS - plugin for InfoLog
 *
 * @link http://www.stylite.de
 * @package stylite
 * @subpackage vfs
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2009-15 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id: Infolog.php 3666 2015-04-07 10:01:31Z ralfbecker $
 */

namespace EGroupware\Stylite\Vfs\Links;

// explicitly import old classes
use infolog_bo;
use translation;

/**
 * VFS plugin for InfoLog
 *
 * This plugin creates virtual subdirectories for InfoLog to allow an easier navigation:
 *
 * - root contains folder for each InfoLog filter (eg. own or own open) and all
 * - next level contains hashes by start date
 *
 * Both public methods (stat and appdir) work by populating stylite_links_stream_wrapper's stat cache
 * using the following methods:
 *
 * - void StreamWrapper::addStatCache(string $path,array|boolean $data);
 * - void StreamWrapper::addStatCacheDir(string $path,string $mime=self::DIR_MIME_TYPE);
 * - void StreamWrapper::addStatCacheLink(string $path,string $link);
 *
 * In general each virtual directory should have an english name implementing the functionality
 * (gets added with addStatCacheDir) and a translated name added as link to english one (addStatCacheLink).
 * As the links get resolved by the vfs layer, the plugin then has only to deal with the english name.
 */
class Infolog implements PluginIface
{
	/**
	 * Use only year as hash, eg. 2009
	 */
	const HASH_YEAR = '%Y';
	/**
	 * Use year and month as hash, eg. 2009-05
	 */
	const HASH_YEAR_MONTH = '%Y-%m';
	/**
	 * Should we hash by year or year and month, use HASH_YEAR or HASH_YEAR_MONTH
	 *
	 * @var string
	 */
	static $hash_format = self::HASH_YEAR;

	/**
	 * Column used for the hash
	 *
	 * @var string
	 */
	static $hash_column = 'info_startdate';

	/**
	 * Instance of infolog
	 *
	 * @var infolog_bo
	 */
	static protected $info;

	/**
	 * Application name of this plugin
	 */
	const APP_NAME = 'infolog';

	/**
	 * Title of addressbook application (translated)
	 *
	 * @var string
	 */
	static protected $app_title;

	/**
	 * Stat something in the application directory
	 *
	 * An implementation can eg. call the appdir method of one level higher, to populate the stat cache
	 * with everything allowed on that level --> StreamWrapper::url_stat then looks up the cache
	 *
	 * @param string $id
	 */
	static public function stat($id)
	{
		//echo "<p>".__METHOD__."($id)</p>\n";

		// we just call appdir to preseed the cache with everything from one level higher
		$pattern = StreamWrapper::virtual_parent($id);

		static $appdir_called_for=null;
		if ($pattern !== $appdir_called_for)
		{
			$appdir_called_for = $pattern;
			self::appdir($pattern);
		}
	}

	/**
	 * Return directory entries for addressbook and populate the stat cache
	 *
	 * - root contains folder for each InfoLog filter (eg. own or own open) and all
	 * - next level contains hashes by start date
	 *
	 * @param string $id
	 * @return array
	 */
	static public function appdir($id)
	{
		if ($id)
		{
			list($label,$hash,$rest) = explode(StreamWrapper::VIRTUAL_DELIMITER,$id);

			if (!($filter = self::label2filter($label)))
			{
				return null;
			}
		}
		//echo '<p>'.__METHOD__."('$id') --> label=$label, hash=$hash --> filter=$filter</p>\n";
		if ($filter && ($hash || $rest === ''))	// --> add entries
		{
			$entries = self::entries($label,$hash,$filter);
		}
		elseif($filter)	// --> add hashes
		{
			$hashes = self::get_hashes($filter);
			if ($hashes['all'] <= StreamWrapper::MAX_ENTRIES_NO_HASH)
			{
				$entries = self::entries($label,'all',$filter);		// show entries without hashes
			}
			else
			{
				$entries = StreamWrapper::hashes2cache($hashes,self::APP_NAME,
					$label.StreamWrapper::VIRTUAL_DELIMITER);
			}
		}
		else	// --> add filters
		{
			foreach(self::$info->filters as $filter => $name)
			{
				if ($filter == 'none') $name = 'all';
				$label = lang($name);
				$link = $name.StreamWrapper::VIRTUAL_DELIMITER;
				StreamWrapper::addStatCacheDir(self::APP_NAME.'/'.$link,'egw/'.self::APP_NAME);
				if ($name != $label)
				{
					StreamWrapper::addStatCacheLink(self::APP_NAME.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,$link);
				}
				if (self::$app_title != self::APP_NAME)
				{
					StreamWrapper::addStatCacheLink(self::$app_title.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,'../'.self::APP_NAME.'/'.$link);

					if ($name != $label) StreamWrapper::addStatCacheLink(self::$app_title.'/'.$link,'../'.self::APP_NAME.'/'.$link);
				}
				$entries[] = $label.StreamWrapper::VIRTUAL_DELIMITER;
			}
		}
		//_debug_array($entries);
		if (StreamWrapper::DEBUG > 1) error_log(__METHOD__."('$id') = ".array2string($entries));
		return $entries;
	}

	/**
	 * Convert a filter label back to a filter
	 *
	 * @param string $label
	 * @return string filter or NULL for unknown labels
	 */
	static private function label2filter($label)
	{
		if ($label == 'all' || $label == lang('all'))
		{
			$filter = 'none';
		}
		elseif (!($filter = array_search($label,self::$info->filters)))
		{
			foreach(self::$info->filters as $f => $name)
			{
				if ($label == lang($name))
				{
					$filter = $f;
					break;
				}
			}
		}
		//echo "<p>".__METHOD__."($label) = ".array2string($filter)."</p>\n";
		return $filter;
	}

	/**
	 * Get the main projects with a given state and evtl. year
	 *
	 * @param string $label filter label
	 * @param string $hash year or year-month or 'all'
	 * @param string $filter filter string of infolog_so
	 * @return array
	 */
	static private function entries($label,$hash,$filter)
	{
		//echo "<p>".__METHOD__."($label,$hash,$filter)</p>\n";
		$entries = array();

		if ($hash == 'all' || $hash == lang('all'))
		{

		}
		elseif(!$hash)	// hash for not set start
		{
			$col_filter[] = self::$hash_column.'=0';
		}
		else
		{
			list($year,$month_start) = explode('-',$hash);
			if (!$month_start)
			{
				$month_start = 1;
				$month_end   = 12;
			}
			else
			{
				$month_end = $month_start;
			}
			$col_filter[] = mktime(0,0,0,$month_start,1,$year).' <= '.self::$hash_column;
			$col_filter[] = self::$hash_column.' < '.mktime(0,0,0,$month_end+1,1,$year);
		}
		foreach(self::$info->search($params=array(
			'filter' => $filter,
			'col_filter' => $col_filter,
			'order' => self::$hash_column,
			'subs' => true,
		)) as $info)
		{
			$id = $info['info_id'];
			$title = ($info[self::$hash_column] ? date('Y-m-d',$info[self::$hash_column]).' ' : '').self::$info->link_title($info);
			$entries[] = $name = StreamWrapper::entry2name(self::APP_NAME,$id,$title);
			StreamWrapper::entry2cache(self::APP_NAME,$id,$name,null,$label.$hash);
		}
		//_debug_array($entries);
		return $entries;
	}

	/**
	 * Get the hash values (eg. creation years) of existing projects of the given status
	 *
	 * @param int $filter
	 * @return array
	 */
	static private function get_hashes($filter)
	{
		static $cache = array();
		$hashes =& $cache[$filter.':'.self::$hash_format];

		if(!isset($hashes))
		{
			//echo "<p>".__METHOD__."($owner) hash_format='".self::$hash_format."' ".function_backtrace()."</p>\n";
			$hashes = array();
			foreach(self::$info->search($params=array(
				'filter' => $filter,
				'cols'  => array(
					$GLOBALS['egw']->db->date_format($GLOBALS['egw']->db->from_unixtime(self::$hash_column),self::$hash_format).' AS hash',
					'COUNT(*) AS num',
				),
				'append' => 'GROUP BY hash',
				'order' => 'hash DESC',
				)) as $row)
			{
				// if more then MAX_ENTRIES_ALL=200 entries in one hash dir, automatic switch to YEAR-MONTH hashes
				if (self::$hash_format == self::HASH_YEAR && $row['num'] > StreamWrapper::MAX_ENTRIES_ALL)
				{
					self::$hash_format = self::HASH_YEAR_MONTH;
					return self::get_hashes($filter);
				}
				if ($row['hash'] == 1970) $row['hash'] = '';
				$hashes[$row['hash']] = $row['num'];
				$total += $row['num'];
			}
			$hashes['all'] = (int)$total;
			//_debug_array($hashes);
		}
		return $hashes;
	}

	/**
	 * Init our static properties
	 *
	 */
	static function init_static()
	{
		self::$info = new infolog_bo();

		// add addressbook translations
		translation::add_app(self::APP_NAME);

		self::$app_title = $GLOBALS['egw_info']['user']['apps'][self::APP_NAME]['title'];
	}
}

Infolog::init_static();
