<?php
/**
 * Stylite: VFS - stream wrapper for linked files
 *
 * @link http://www.stylite.de
 * @package stylite
 * @subpackage links
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2008-15 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id: StreamWrapper.php 3666 2015-04-07 10:01:31Z ralfbecker $
 */

/**
 * Use Stylite\Vfs\Versioning\StreamWrapper as parent for the Api\Vfs\Links\StreamWrapper to enable versioning
 */
namespace EGroupware\Api\Vfs\Links;

if (!class_exists('EGroupware\\Api\\Vfs\\Links\\LinksParent', false))
{
	class LinksParent extends \EGroupware\Stylite\Vfs\Versioning\StreamWrapper {}
}

/**
 * Our own namespace in Stylite app
 */
namespace EGroupware\Stylite\Vfs\Links;

use EGroupware\Api\Vfs;

// explicitly import old phpgwapi classes used:
use egw_link;
use egw_cache;

/**
 * Stylite: stream wrapper for linked files
 *
 * The files stored by the Vfs\Sqlfs\StreamWrapper in a /apps/$app/$id directory
 *
 * The links stream wrapper extends the sqlfs one, to implement an own ACL based on the access
 * of the entry the files are linked to.
 *
 * Applications can define a 'file_access' method in the link registry with the following signature:
 *
 * 		boolean function file_access(string $id,int $check,string $rel_path)
 *
 * If the do not implement such a function the title function is used to test if the user has
 * at least read access to an entry, and if true full (write) access to the files is granted.
 *
 * The stream wrapper interface is according to the docu on php.net
 *
 * @link http://www.php.net/manual/en/function.stream-wrapper-register.php
 *
 * @todo several methods need to clear the stat-cache!
 */
class StreamWrapper extends Vfs\Links\StreamWrapper
{
	/**
	 * Scheme / protocoll used for this stream-wrapper
	 */
	const SCHEME = 'stylite.links';
	/**
	 * Prefix to predend to get an url from a path
	 */
	const PREFIX = 'stylite.links://default';
	/**
	 * Base url to applicaton data
	 */
	const BASEURL = 'stylite.links://default/apps';
	/**
	 * Base path to application data
	 */
	const BASEPATH = '/apps';

	/**
	 * Delimiter for virtual directories (trailing character)
	 *
	 * Windows does NOT allow an asterisk (*)!
	 */
	const VIRTUAL_DELIMITER = '$';

	/**
	 * Maximum number of entries displayed without hashing (virtual directory level with first letter)
	 */
	const MAX_ENTRIES_NO_HASH = 100;

	/**
	 * Maximum number of entries to display an all$ directory
	 *
	 */
	const MAX_ENTRIES_ALL = 200;

	/**
	 * How much should be logged to the apache error-log
	 *
	 * 0 = Nothing
	 * 1 = only errors
	 * 2 = all function calls and errors (contains passwords too!)
	 * 3 = more diagnostics
	 */
	const LOG_LEVEL = 1;

	/**
	 * Application of the opened dir
	 *
	 * @var string
	 */
	protected $dir_app;
	/**
	 * Application id of the opened dir
	 *
	 * @var int|string
	 */
	protected $dir_id;
	/**
	 * Subdir name of the opened dir
	 *
	 * @var string
	 */
	protected $dir_rest;
	/**
	 * Links of the entry, whos dir is opened
	 *
	 * @var array
	 */
	protected $dir_links;

	/**
	 * Directory vfs::ls() of dir opened with dir_opendir()
	 *
	 * @var array $path => stat pairs
	 */
	static protected $stat_cache = array();

	/**
	 * Shortcuts for application names, the other apps get added in init_static()
	 *
	 * @var array
	 */
	static public $shortcut2app = array(
		'A' => 'addressbook',
		'C' => 'calendar',
		'H' => 'timesheet',
		'I' => 'infolog',
		'K' => 'phpbrain',
		'P' => 'projectmanager',
		'R' => 'resources',
		'T' => 'tracker',
	);

	/**
	 * This method is called immediately when your stream object is created for examining directory contents with opendir().
	 *
	 * @param string $url URL that was passed to opendir() and that this object is expected to explore.
	 * @param $options
	 * @return booelan
	 */
	function dir_opendir ( $url, $options )
	{
		$path = Vfs::parse_url($url,PHP_URL_PATH);

		parse_str(parse_url($url, PHP_URL_QUERY), $this->dir_url_params);
		
		$app = $id = $rest = null;
		if (!self::path2entry($path,$app,$id,$rest) && !(is_null($id) || substr($id,-1) == self::VIRTUAL_DELIMITER))
		{
			return false;
		}
		$this->dir_app  = $app;
		$this->dir_id   = $id;
		$this->dir_rest = $rest;

		// applications directory --> return all apps participating in linking the user has rights to
		if (empty($app))
		{
			if (self::LOG_LEVEL > 2) error_log(__METHOD__."($path,$options) app=$app, id=$id, rest=$rest) applications dir");
			$this->opened_dir = array();

			// adding favorites like an application
			$app = 'favorites';
			// we dont show the favorites, until the symlink problem is solved!
			//$this->opened_dir[] = $title = lang('Favorites');
			//self::app2cache($app,$title);

			foreach(self::$shortcut2app as $app)
			{
				if (!isset($GLOBALS['egw_info']['user']['apps'][$app])) continue;	// ignore apps the user has no rights
				$this->opened_dir[] = $title = $GLOBALS['egw_info']['user']['apps'][$app]['title'];
				self::app2cache($app,$title);
			}
			$ret = true;
		}
		// application directory --> always exist, if user has app-rigths and returns ALL entries
		elseif((is_null($id) || substr($id,-1) == self::VIRTUAL_DELIMITER) && ($app_class = self::app_class($app)))
		{
			$this->opened_dir = call_user_func(array($app_class, 'appdir'), $id, self::url_params($url));
			$ret = $this->opened_dir !== false;
		}
		// application directory --> always exist, if user has app-rigths and returns ALL entries
		elseif(is_null($id))
		{
			if (self::LOG_LEVEL > 2) error_log(__METHOD__."($path,$options) app=$app, id=$id, rest=$rest) app-dir");

			$this->opened_dir = self::appdir($app);
			$ret = $this->opened_dir !== false;
		}
		// entry directory --> always exists, if there's the coresponding entry (even if not existing in vfs!)
		elseif(is_null($rest))
		{
			if (self::LOG_LEVEL > 2) error_log(__METHOD__."($path,$options) app=$app, id=$id, rest=$rest) entry-dir");

			if (!parent::url_stat($url,STREAM_URL_STAT_QUIET) || !($ret = parent::dir_opendir($url,$options)))
			{
				$this->opened_dir = array();
				$ret = true;
			}
		}
		// subdir in entry-directory --> pass directly to parent
		else
		{
			if (self::LOG_LEVEL > 2) error_log(__METHOD__."($path,$options) app=$app, id=$id, rest=$rest) subdir of entry-dir");

			$ret = parent::dir_opendir($url,$options);
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($path,$options) app=$this->dir_app, id=$this->dir_id, rest=$this->dir_rest) returning ".array2string($ret).', opened_dir='.array2string($this->opened_dir));
		return $ret;
	}

	/**
	 * Load application specific class either via autoloading or from stylite class directory
	 *
	 * @param string $app application name or title(!)
	 * @return string|boolean classname or false if not found
	 */
	static function app_class($app)
	{
		static $title2app = array();

		if ($app == 'favorites' || $app == lang('Favorites'))
		{
			$app = 'favorites';
			$class = '\stylite_filemanager_favorites';
		}
		elseif (!isset($GLOBALS['egw_info']['user']['apps'][$app]))
		{
			if (!isset($title2app[$app]))
			{
				foreach($GLOBALS['egw_info']['user']['apps'] as $name => $data)
				{
					if ($data['title'] == $app)
					{
						$title2app[$app] = $name;
						break;
					}
				}
			}
			if (!isset($title2app[$app]))
			{
				return false;
			}
			$app = $title2app[$app];
		}
		if ((isset($class) ||
			class_exists($class = __NAMESPACE__.'\\'.ucfirst($app)) ||	// new name in EGroupware\Stylite\Vfs\Links
			class_exists($class = '\\'.$app.'_vfs')) &&	// old name in app itself
			is_subclass_of($class, __NAMESPACE__.'\\PluginIface', true))
		{
			return $class;
		}
		return false;
	}

	/**
	 * Return $app's entries and add them to the cache
	 *
	 * @param string $app
	 * @param string $basepath =null default $app
	 * @return array|boolean false on error (eg. no app rights) or array with file-names
	 */
	static function appdir($app,$basepath=null)
	{
		if (!isset($GLOBALS['egw_info']['user']['apps'][$app]))
		{
			$title = $app;
			// check if $title matches an app-title
			foreach($GLOBALS['egw_info']['user']['apps'] as $app => $data)
			{
				if ($data['title'] === $title) break;
			}
			if ($data['title'] !== $title) return false;
		}
		$entries = array();
		foreach(egw_link::query($app,'') as $id => $title)
		{
			if (is_array($title)) $title = $title['lable'];
			$entries[] = $name = self::entry2name($app,$id,$title);
			self::entry2cache($app,$id,$name,null,$basepath);
		}
		return $entries;
	}

	/**
	 * Add an alphabetic selection to the cache
	 *
	 * @param string $app
	 * @param string $prefix eg. "accounts*"
	 * @param string $prefix_alias eg. "Benutzerkonten*"
	 * @param array  $alphabet =null array with letter => number of matches pairs, or null to use standard alphabet without number of matches
	 * @return array
	 */
	static function alphabet2cache($app,$prefix,$prefix_alias=null,array $alphabet=null)
	{
		if (is_null($alphabet))
		{
			if (substr($lang_alphabet = lang('alphabet'), 8) == 'alphabet')
			{
				$lang_alphabet = 'a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z';
			}
			$alphabet_arr = explode(',',strtoupper($lang_alphabet));
			$alphabet = array_combine($alphabet_arr,array_fill(0,count($alphabet_arr),null));
			$alphabet['all'] = '';
		}
		return self::hashes2cache($alphabet,$app,$prefix,$prefix_alias);
	}

	/**
	 * Add hashed selection to the cache
	 *
	 * Displays number of entries in $hashes --> size of the dir
	 * If hash contains year[-month] add that as modification date to allow sorting by it
	 *
	 * @param array  $hashes array with hash (eg. letter) => number of matches pairs, or null to use standard alphabet without number of matches
	 * @param string $app
	 * @param string $prefix eg. "accounts*"
	 * @param string $prefix_alias eg. "Benutzerkonten*"
	 * @return array
	 */
	static function hashes2cache(array $hashes,$app,$prefix,$prefix_alias=null)
	{
		if (is_null($prefix_alias) && ($stat = self::$stat_cache[self::BASEPATH.'/'.$app.'/'.$prefix]) && $stat['readlink'])
		{
			$prefix_alias = $prefix;
			$prefix = Vfs::basename($stat['readlink']);
		}
		$entries = array();
		foreach($hashes as $hash => $num)
		{
			$ts = $matches = null;
			// check if hash contains year[-month] --> use that as timestamp, to allow sorting by mtime
			if (ctype_digit(substr($hash,-1)) && preg_match('/-?([0-9]{4})(-([0-9]{2}))?$/',$hash,$matches))
			{
				$ts = mktime(0,0,0,isset($matches[3])?(int)$matches[3]:1,1,(int)$matches[1]);
			}
			$name = $label = $hash.self::VIRTUAL_DELIMITER;
			if ($hash == 'all')
			{
				if ($num > self::MAX_ENTRIES_ALL) continue;		// to many entries --> dont display all link
				$label = lang('all').self::VIRTUAL_DELIMITER;	// translate all
			}
			$entries[] = $label;
			$link = !empty($prefix) ? '../'.$prefix.$name : $name;

			self::addStatCacheDir($app.'/'.$prefix.$name,'egw/'.$app,$ts,$num);

			if ($name != $label)
			{
				self::addStatCacheLink($app.'/'.(empty($prefix) ? '' : $prefix.'/').$label,$link,$ts,$num);
			}
			if (!empty($prefix)) self::addStatCacheLink($app.'/'.$prefix.'/'.$name,$link,$ts,$num);

			if (!is_null($prefix_alias) && $prefix_alias != $prefix)
			{
				self::addStatCacheLink($app.'/'.$prefix_alias.'/'.$name,$link,$ts,$num);
			}
		}
		return $entries;
	}

	/**
	 * Add some data to the stat cache
	 *
	 * @param string $path relative to BASEPATH
	 * @param array|boolean $data array with stat-data or false
	 */
	static public function addStatCache($path,$data)
	{
		//echo "<p>".__METHOD__."($path,".array2string($data).')'.function_backtrace()."</p>\n";
		if ($path[0] != '/') $path = self::BASEPATH.'/'.$path;

		self::$stat_cache[$path] = $data;
	}

	/**
	 * Adding a directory to the stat cache
	 *
	 * @param string $path
	 * @param string $mime =self::DIR_MIME_TYPE
	 * @param int $ts =null timestamp for m/ctime, default null = current time
	 * @param int $size =0 size
	 */
	static public function addStatCacheDir($path,$mime=self::DIR_MIME_TYPE,$ts=null,$size=0)
	{
		//echo "<p>".__METHOD__."($path,$mime,$ts,$size) ".function_backtrace()."</p>\n";
		if (is_null($ts)) $ts = time();

		self::addStatCache($path,array(
			'ino'   => md5($path),
			'name'  => Vfs::basename($path),
			'mode'  => 05 | self::MODE_DIR,
			'size'  => $size,
			'uid'   => 0,
			'gid'   => 0,
			'mtime' => $ts,
			'ctime' => $ts,
			'nlink' => 2,
			'mime'  => $mime,
		));
	}

	/**
	 * Adding a Link to the stat cache
	 *
	 * @param string $path
	 * @param string $link
	 * @param int $ts =null timestamp for m/ctime, default null = current time
	 * @param int $size =null size
	 */
	static public function addStatCacheLink($path,$link,$ts=null,$size=null)
	{
		//echo "<p>".__METHOD__."($path,$link,$ts,$size) ".function_backtrace()."</p>\n";
		if (is_null($ts)) $ts = time();

		self::addStatCache($path,array(
			'ino'   => md5($path),
			'name'  => Vfs::basename($path),
			'mode'  => 04 | self::MODE_LINK,
			'size'  => !is_null($size) ? $size : bytes($link),
			'uid'   => 0,
			'gid'   => 0,
			'mtime' => $ts,
			'ctime' => $ts,
			'nlink' => 2,
			'mime'  => self::SYMLINK_MIME_TYPE,
			'readlink' => $link,
		));
	}

	/**
	 * Return the stat cache
	 *
	 * @param string $path =null null to get whole stat cache or path to get only data for $path
	 * @return array|boolean|NULL
	 */
	static public function getStatCache($path=null)
	{
		return is_null($path) ? self::$stat_cache : self::$stat_cache[$path];
	}

	/**
	 * Return the stat cache
	 *
	 * @param string $path =null null to get whole stat cache or path to get only data for $path
	 * @todo also delete stat-cache entries which symlink to deleted entry
	 *	eg. /apps/InfoLog/.../...(I--123)/something when /apps/infolog/123/something should be deleted
	 */
	static public function deleteStatCache($path=null)
	{
		if ($path[0] != '/') $path = Vfs::parse_url($path,PHP_URL_PATH);

		//echo "<p>".__METHOD__."($path)".function_backtrace()."</p>\n";
		unset(self::$stat_cache[$path]);

		/* this code deletes above symlinked entry, but is not sufficient
		$app = $id = $rest = $link = null;
		if (self::$stat_cache && self::path2entry($path, $app, $id, $rest, $link))
		{
			foreach(array_keys(self::$stat_cache) as $p)
			{
				$a = $i = $r = $l = null;
				if (self::path2entry($p, $a, $i, $r, $l) &&
					($l == $link || $l == $path || $a == $app && $i == $id && $r == $rest))
				{
					error_log(__METHOD__."('$path') also deleting $p");
				}
			}
		}*/
		// for now we clear whole stat-cache
		self::$stat_cache = array();
	}

	/**
	 * Clears our stat-cache
	 *
	 * Normaly not necessary, as it is automatically cleared/updated, UNLESS Vfs::$user changes!
	 * We have to clear everything down to the entry directory eg. /apps/tracker/123, as it's traversal rights may change.
	 * Currently we clear the whole cache, to work around that.
	 *
	 * @todo clear only from $path down to entry dir (inclusive), but not everything
	 * @param string $path ='/'
	 */
	static public function clearstatcache($path)
	{
		//error_log(__METHOD__."('$path')");
		self::$stat_cache = array();

		parent::clearstatcache($path);
	}

	/**
	 * Add an application entry (incl. the title symlink) to stat cache
	 *
	 * @param string $app
	 * @param string|int $id
	 * @param string $name =null filename (title [app:id])
	 * @param string $link =null symlink, default $id
	 * @param string $basepath =null default $app
	 */
	static function entry2cache($app,$id,$name=null,$link=null,$basepath=null)
	{
		if (is_null($name)) $name = self::entry2name($app,$id);
		if (is_null($basepath)) $basepath = $app;
		if (is_null($link)) $link = $basepath == $app ? $id : '../'.$app.'/'.$id;

		self::addStatCacheDir($app.'/'.$id,'egw/'.$app);
		self::addStatCacheLink($basepath.'/'.$name,$link);
	}

	/**
	 * Add an application directory (and the title symlink) to stat cache
	 *
	 * @param string $app
	 * @param string $title =null
	 * @return boolean false on error (eg. no app rights), true otherwise
	 */
	static function app2cache($app,$title=null)
	{
		if ($app == 'favorites' || $app == lang('Favorites'))
		{
			$app = 'favorites';
			$title = lang('Favorites');
		}
		elseif (!isset($GLOBALS['egw_info']['user']['apps'][$app]))
		{
			// check if $title matches an app-title
			foreach($GLOBALS['egw_info']['user']['apps'] as $app => $data)
			{
				if ($data['title'] === $title) break;
			}
			if ($data['title'] !== $title) return false;
		}
		if (is_null($title)) $title = $GLOBALS['egw_info']['user']['apps'][$app]['title'];

		self::addStatCacheDir($app,'egw/'.$app);

		if ($title != $app)
		{
			self::addStatCacheLink($title,$app);
		}
		return true;
	}

	/**
	 * This method is called in response to readdir().
	 *
	 * It should return a string representing the next filename in the location opened by dir_opendir().
	 *
	 * @return string
	 */
	function dir_readdir ( )
	{
		// are we reading the content of an app directory
		if (($file = parent::dir_readdir()) !== false && $this->dir_app && is_null($this->dir_id) && is_null($this->dir_rest))
		{

		}
		// are we finished (!$file) reading the dir of an entry
		elseif($file === false && !is_null($this->dir_id) &&
			is_null($this->dir_rest) && !$this->dir_url_params['backup'])
		{
			if (is_null($this->dir_links))
			{
				$this->dir_links = egw_link::get_links($this->dir_app,$this->dir_id,'!'.egw_link::VFS_APPNAME,'link_lastmod DESC',true);
				reset($this->dir_links);
			}
			if (($link = current($this->dir_links)))
			{
				next($this->dir_links);
				$file = self::entry2name($link['app'],$link['id']);
			}
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."( ) app=$this->dir_app, id=$this->dir_id, rest=$this->dir_rest --> ".array2string($file));
		return $file;
	}

	/**
	 * This method is called in response to rewinddir().
	 *
	 * Reimplemented to reset $this->dir_links too.
	 *
	 * It should reset the output generated by dir_readdir(). i.e.:
	 * The next call to dir_readdir() should return the first entry in the location returned by dir_opendir().
	 *
	 * @return boolean
	 */
	function dir_rewinddir ( )
	{
		if (!parent::dir_rewinddir()) return false;

		if (!is_null($this->dir_links)) reset($this->dir_links);

		return true;
	}

	/**
	 * Get the name of an entry: build by title postfixed with [app:id]
	 *
	 * @param string $app
	 * @param string|int $id
	 * @return string|boolean name or false if permission denied or NULL if not found, see egw_link::title()
	 */
	static function entry2name($app,$id,$title=null)
	{
		if (is_null($title)) $title = egw_link::title($app,$id);
		$name = $title;
		if ($name !== false && !is_null($name))
		{
			if (!is_string($name) && self::LOG_LEVEL > 0) error_log(__METHOD__."($app,$id): egw_link::title($app,$id) = ".array2string($name));

			$name = self::sanitize_filename($name);	//remove or replace not allowed chars

			// Remove ID, if included in title (Infolog, maybe others)
			if (true) $name = str_replace('('.$id.')', '',$name);

			if (!($shortcut = array_search($app,self::$shortcut2app))) $shortcut = $app;
			$name .= ' ('.$shortcut.'--'.$id.')';
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($app,$id) --> ".array2string($name));

		return $name;
	}

	/**
	 * Sanitize a filename (remove or replace not allowed chars)
	 *
	 * @param string $name
	 * @return string
	 */
	static function sanitize_filename($name)
	{
		return strtr($name,array(
		// we can not allow path delimiters in the name --> replacing them with a dash (-)
			'/' => '-',
			'\\'=> '-',
		// replace some chars which invalid in filenames for windows (:)
			':' => ',',
		// remove some chars which mess up url parsing
			'#' => '',
			'?' => '',
		));
	}

	const APP_ID_PREG = '/ \(([A-Za-z]+)--([^\)]+)\)(\/.*?|)$/U';

	/**
	 * Parse a path (with [app:id]) and extract $app, $id and evtl. a rest
	 *
	 * @param string $path
	 * @param string &$app
	 * @param string &$id
	 * @param string &$rest=null
	 * @param string &$link=null
	 * @return boolean true if parsable, false otherwise
	 */
	static function path2entry($path,&$app,&$id,&$rest=null,&$link=null)
	{
		$app = $id = $rest = $matches = null;
		if (($ret = preg_match(self::APP_ID_PREG,$path,$matches)))
		{
			list(,$app,$id,$rest) = $matches;
			// use the last occurence (unfortunately U=Ungreedy modifier does not work in that case)
			while($rest && preg_match(self::APP_ID_PREG,$rest,$matches))
			{
				list(,$app,$id,$rest) = $matches;
			}
			if (isset(self::$shortcut2app[strtoupper($app)])) $app = self::$shortcut2app[strtoupper($app)];
			if ($rest && $rest[0] == '/') $rest = substr($rest,1);
			$link = egw_link::vfs_path($app,$id,$rest,true);
		}
		elseif((list(,,$app,$id,$rest) = explode('/',$path,5)) && isset($GLOBALS['egw_info']['user']['apps'][$app]))
		{
			$link = $path;
			$ret =  !empty($id) && substr($id,-1) != self::VIRTUAL_DELIMITER;
		}
		if ($ret) $ret = parent::check_extended_acl($link,Vfs::READABLE);

		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($path) --> app=$app, id=$id, rest=$rest, returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Geting the pattern of the parent level (of the virtual hierarchy):
	 *
	 * "hugo$" --> ""
	 * "hugo$2009$" --> "hugo$"
	 *
	 * @param string $id
	 * @return string
	 */
	static function virtual_parent($id)
	{
		// getting the $pattern for one level higher
		$parts = explode(self::VIRTUAL_DELIMITER,$id);
		array_pop($parts);
		array_pop($parts);
		if ($parts) $parts[] = '';	// if something left append a $

		return implode(self::VIRTUAL_DELIMITER,$parts);
	}

	/**
	 * Return parsed URL / GET parameters as array
	 *
	 * @param string $url
	 * @return array
	 */
	static function url_params($url)
	{
		$params = array();
		if (($query = Vfs::parse_url($url, PHP_URL_QUERY)))
		{
			parse_str($query, $params);
		}
		return $params;
	}

	/**
	 * This method is called in response to stat() calls on the URL paths associated with the wrapper.
	 *
	 * Reimplemented from links to check of app:id in path and return a symlink if it's the case.
	 *
	 * @param string $url
	 * @param int $flags holds additional flags set by the streams API. It can hold one or more of the following values OR'd together:
	 * - STREAM_URL_STAT_LINK	For resources with the ability to link to other resource (such as an HTTP Location: forward,
	 *                          or a filesystem symlink). This flag specified that only information about the link itself should be returned,
	 *                          not the resource pointed to by the link.
	 *                          This flag is set in response to calls to lstat(), is_link(), or filetype().
	 * - STREAM_URL_STAT_QUIET	If this flag is set, your wrapper should not raise any errors. If this flag is not set,
	 *                          you are responsible for reporting errors using the trigger_error() function during stating of the path.
	 *                          stat triggers it's own warning anyway, so it makes no sense to trigger one by our stream-wrapper!
	 * @return array
	 */
	static function url_stat ( $url, $flags )
	{
		$path = Vfs::parse_url($url,PHP_URL_PATH);

		$app = $id = $rest = $link = null;
		if (isset(self::$stat_cache[$path]))
		{
			$mode = 'from stat_cache';
		}
		elseif(($stat = parent::url_stat($url,$flags)))
		{
			self::addStatCache($path,$stat);
			$mode = 'parent::url_stat()';
		}
		elseif(self::path2entry($path,$app,$id,$rest,$link) && (is_null($rest) || $rest === ''))
		{
			self::addStatCacheLink($path,$link);
			self::addStatCacheDir($link,'egw/'.$app);
			$mode = 'own symlinks';
		}
		elseif($app && is_null($id))
		{
			if (!self::app2cache($app,$app)) self::$stat_cache[$path] = false;
			$mode = 'app directory';
		}
		elseif($app && (is_null($rest) || substr($rest,-1) == self::VIRTUAL_DELIMITER) &&
			(substr($id,-1) == self::VIRTUAL_DELIMITER || $app == 'favorites') &&
			($app_class = self::app_class($app)))
		{
			call_user_func(array($app_class,'stat'), $id.$rest, self::url_params($url));
			$mode = $app_class.'::stat';
		}
		else
		{
			self::addStatCache($path,false);
			$mode = '!parent::url_stat()';
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url): $mode --> ".array2string(self::$stat_cache[$path]));
		return isset(self::$stat_cache[$path]) ? self::$stat_cache[$path] : false;
	}

	/**
	 * This method is called in response to readlink().
	 *
	 * The readlink value is read by url_stat or dir_opendir and therefore cached in the stat-cache.
	 *
	 * Reimplemented from Vfs\Sqlfs\StreamWrapper, as we have no late static binding and we need our url_stat.
	 *
	 * @param string $path
	 * @return string|boolean content of the symlink or false if $url is no symlink (or not found)
	 */
	static function readlink($path)
	{
		$link = !($lstat = self::url_stat($path,STREAM_URL_STAT_LINK)) || is_null($lstat['readlink']) ? false : $lstat['readlink'];

		if (self::LOG_LEVEL > 1) error_log(__METHOD__."('$path') = $link");

		return $link;
	}

	/**
	 * Implements ACL based on the access of the user to the entry the files are linked to.
	 *
	 * Reimplemented to be to "understand" [app:id]
	 *
	 * @param string $url url to check
	 * @param int $check mode to check: one or more or'ed together of: 4 = read, 2 = write, 1 = executable
	 * @return boolean
	 */
	static function check_extended_acl($url,$check)
	{
		$path = $url[0] == '/' ? $url : Vfs::parse_url($url,PHP_URL_PATH);
		$app = $id = $rest = null;
		if (self::path2entry($path,$app,$id,$rest))
		{
			$path = egw_link::vfs_path($app,$id,$rest,true);
		}
		return parent::check_extended_acl($path,$check);
	}

	/**
	 * This method is called immediately after your stream object is created.
	 *
	 * Reimplemented to create entry directories on the fly, as they are always displayed as existing.
	 *
	 * @param string $url URL that was passed to fopen() and that this object is expected to retrieve
	 * @param string $mode mode used to open the file, as detailed for fopen()
	 * @param int $options additional flags set by the streams API (or'ed together):
	 * - STREAM_USE_PATH      If path is relative, search for the resource using the include_path.
	 * - STREAM_REPORT_ERRORS If this flag is set, you are responsible for raising errors using trigger_error() during opening of the stream.
	 *                        If this flag is not set, you should not raise any errors.
	 * @param string $opened_path full path of the file/resource, if the open was successfull and STREAM_USE_PATH was set
	 * @return boolean true if the ressource was opened successful, otherwise false
	 */
	function stream_open ( $url, $mode, $options, &$opened_path )
	{
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url,$mode,$options)");

		// the following call is necessary to fill Vfs\Sqlfs\StreamWrapper::$stat_cache, WITH the extendes ACL!
		//parent::url_stat($url,0);	// --> moved to links_stream_wrapper::stream_open()

		if ($mode[0] != 'r' && !parent::url_stat($dir = Vfs::dirname($url),0) && parent::check_extended_acl($dir,Vfs::WRITABLE))
		{
			parent::mkdir($dir,0,STREAM_MKDIR_RECURSIVE);
		}
		self::deleteStatCache($url);
		return parent::stream_open($url,$mode,$options,$opened_path);
	}

	/**
	 * This method is called when the stream is closed, using fclose().
	 *
	 * Reimplemented to clear our stat-cache.
	 */
	function stream_close ( )
	{
		self::deleteStatCache($this->opened_path);

		parent::stream_close();
	}

	/**
	 * Method called for symlink()
	 *
	 * Reimplemented to create an entry directory on the fly AND delete our stat cache!
	 *
	 * @param string $target
	 * @param string $link
	 * @return boolean true on success false on error
	 */
	static function symlink($target,$link)
	{
		if (!parent::url_stat($dir = Vfs::dirname($link),0) && parent::check_extended_acl($dir,Vfs::WRITABLE))
		{
			parent::mkdir($dir,0,STREAM_MKDIR_RECURSIVE);
		}
		self::deleteStatCache($link);
		return parent::symlink($target,$link);
	}

	/**
	 * This is not (yet) a stream-wrapper function, but it's necessary and can be used static
	 *
	 * Reimplemented to create an entry directory on the fly AND delete our stat cache!
	 *
	 * @param string $url
	 * @param int $time =null modification time (unix timestamp), default null = current time
	 * @param int $atime =null access time (unix timestamp), default null = current time, not implemented in the vfs!
	 */
	static function touch($url,$time=null,$atime=null)
	{
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url,$time,$atime)");

 		if (!($stat = self::url_stat($url,STREAM_URL_STAT_QUIET)))
		{
			// file does not exist --> create an empty one
			if (!($f = fopen(self::SCHEME.'://default'.Vfs::parse_url($url,PHP_URL_PATH),'w')) || !fclose($f))
			{
				return false;
			}
		}

		self::deleteStatCache($url);
		return is_null($time) ? true : parent::touch($url,$time,$atime);
	}

	/**
	 * Read properties for a ressource (file, dir or all files of a dir)
	 *
	 * Reimplemented to call parent only on real dirs not virtual or not (yet) existing ones
	 *
	 * @param array|string $pathes (array of) string with path
	 * @param string $ns ='http://egroupware.org/' namespace if propfind should be limited to a single one, otherwise use null
	 * @return array|boolean array with props (values for keys 'name', 'ns', 'val'), or path => array of props for is_array($path)
	 * 	false if $path does not exist
	 */
	static function propfind($pathes,$ns=self::DEFAULT_PROP_NAMESPACE)
	{
		foreach((array)$pathes as $key => $path)
		{
			if ($path[0] != '/') $path = Vfs::parse_url($path,PHP_URL_PATH);
			list(,$apps,$app,$id) = explode('/',$path);

			if ($apps != 'apps' || !isset($GLOBALS['egw_info']['user']['apps'][$app]) || !is_numeric($id) ||
				!parent::url_stat($path,STREAM_URL_STAT_QUIET))
			{
				if (!is_array($pathes)) return false;

				unset($pathes[$key]);
			}
		}
		if (!$pathes) return false;	// no existing path left

		return parent::propfind($pathes,$ns);
	}

	/**
	 * This method is called in response to mkdir() calls on URL paths associated with the wrapper.
	 *
	 * Reimplemented to clear our stat-cache
	 *
	 * @param string $path
	 * @param int $mode not used(!), we inherit 005 for /apps/$app and set 000 for /apps/$app/$id
	 * @param int $options Posible values include STREAM_REPORT_ERRORS and STREAM_MKDIR_RECURSIVE, we allways use recursive!
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function mkdir($path,$mode,$options)
	{
		self::deleteStatCache($path);

		return parent::mkdir($path,$mode,$options);
	}

	/**
	 * This method is called in response to rmdir() calls on URL paths associated with the wrapper.
	 *
	 * Reimplemented to do nothing (specially not complain), if an entry directory does not exist,
	 * as we always report them as existing!
	 * We never delete entry directories, as this makes there attic inaccessible
	 * (nothing to gain here anyway, as database entry is never deleted in versioning).
	 *
	 * @param string $url
	 * @param int $options Possible values include STREAM_REPORT_ERRORS.
	 * @return boolean TRUE on success or FALSE on failure.
	 */
	static function rmdir ( $url, $options )
	{
		$path = $url != '/' ? Vfs::parse_url($url,PHP_URL_PATH) : $url;

		list(,/*$apps*/,/*$app*/,/*$id*/,$rest) = explode('/',$path);

		// never delete entry-dir, as it makes attic inaccessible
		if (empty($rest))
		{
			return true;
		}
		return parent::rmdir( $path, $options );
	}

	/**
	 * This method is called in response to rename() calls on URL paths associated with the wrapper.
	 *
	 * Reimplemented to clear our stat-cache
	 *
	 * @param string $path_from
	 * @param string $path_to
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function rename ( $path_from, $path_to )
	{
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($path_from,$path_to)");

		self::deleteStatCache($path_from);
		self::deleteStatCache($path_to);

		return parent::rename($path_from,$path_to);
	}

	/**
	 * Init our static properties
	 */
	static public function init_static()
	{
		if (($shortcuts = egw_cache::getSession(__CLASS__,'shortcut2app')))
		{
			self::$shortcut2app = $shortcuts;
		}
		else
		{
			foreach(egw_link::app_list('query') as $app => $title)
			{
				unset($title);
				if (in_array($app,array('mydms','wiki'))) continue;	// they do NOT support adding files

				if (!in_array($app,self::$shortcut2app))
				{
					$long = strtoupper(str_replace('perp','p',str_replace(array('-','_'),'',$app)));
					for($n = 1; $n <= strlen($app); ++$n)
					{
						if (!isset(self::$shortcut2app[$shortcut=substr($long,0,$n)]))
						{
							self::$shortcut2app[$shortcut] = $app;
							break;
						}
					}
				}
			}
			//_debug_array(self::$shortcut2app);
			egw_cache::setSession(__CLASS__,'shortcut2app',self::$shortcut2app);
		}
		// register our scheme
		stream_register_wrapper(self::SCHEME ,__CLASS__);
	}

	/**
	 * Check and optionaly fix corruption: undelete entry dirs, to make their attic accessible
	 *
	 * @param array $params values for keys 'location' and 'check_only'
	 * @return array with messages / found problems
	 */
	public static function fsck($params)
	{
		error_log(__METHOD__.'('.array2string($params).')');
		if (!($apps = self::url_stat('/apps', 0))) return array();

		$msgs = array();
		$app_dir_ids = 'SELECT fs_id FROM '.self::TABLE.' WHERE fs_dir='.(int)$apps['ino'];
		$ids = self::$pdo->query('SELECT fs_id FROM '.self::TABLE.
			" WHERE fs_mime='httpd/unix-directory' AND fs_active=".self::_pdo_boolean(false).
			" AND fs_dir IN ($app_dir_ids)")->fetchAll(\PDO::FETCH_COLUMN, 0);

		if (count($ids))
		{
			if (is_array($params) && isset($params['check_only']) && $params['check_only'])
			{
				$msgs[] = lang('%1 attachment entry directories need to be fixed in order to access their deleted files', count($ids));
			}
			else
			{
				$rs = self::$pdo->query('UPDATE '.self::TABLE.' SET fs_active='.self::_pdo_boolean(true).
					' WHERE fs_id IN ('.implode(',', $ids).')');
				$msgs[] = lang('%1 attachment directories fixed', $rs->rowCount());
			}
		}
		return $msgs;
	}
}

StreamWrapper::init_static();
