<?php
/**
 * Stylite: VFS - plugin for tracker
 *
 * @link http://www.stylite.de
 * @package stylite
 * @subpackage vfs
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2009-15 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id: Tracker.php 3666 2015-04-07 10:01:31Z ralfbecker $
 */

namespace EGroupware\Stylite\Vfs\Links;

// explicitly import old classes
use tracker_bo;
use translation;

/**
 * VFS plugin for tracker
 *
 * This plugin creates virtual subdirectories for tracker to allow an easier navigation:
 *
 * - root contains folder for each tracker queue
 * - next level contains open tickets (or hashes starting with lang('open')) and dirs for status pending and closed (evtl with hashes)
 *
 * Both public methods (stat and appdir) work by populating stylite_links_stream_wrapper's stat cache
 * using the following methods:
 *
 * - void StreamWrapper::addStatCache(string $path,array|boolean $data);
 * - void StreamWrapper::addStatCacheDir(string $path,string $mime=self::DIR_MIME_TYPE);
 * - void StreamWrapper::addStatCacheLink(string $path,string $link);
 *
 * In general each virtual directory should have an english name implementing the functionality
 * (gets added with addStatCacheDir) and a translated name added as link to english one (addStatCacheLink).
 * As the links get resolved by the vfs layer, the plugin then has only to deal with the english name.
 */
class Tracker implements PluginIface
{
	/**
	 * Use only year as hash, eg. 2009
	 */
	const HASH_YEAR = '%Y';
	/**
	 * Use year and month as hash, eg. 2009-05
	 */
	const HASH_YEAR_MONTH = '%Y-%m';
	/**
	 * Should we hash by year or year and month, use HASH_YEAR or HASH_YEAR_MONTH
	 *
	 * @var string
	 */
	static $hash_format = self::HASH_YEAR;

	/**
	 * Column used for the hash
	 *
	 * @var string
	 */
	static $hash_column = 'tr_created';

	/**
	 * Instance of tracker
	 *
	 * @var tracker_bo
	 */
	static protected $tracker;

	/**
	 * Application name of this plugin
	 */
	const APP_NAME = 'tracker';

	/**
	 * Title of addressbook application (translated)
	 *
	 * @var string
	 */
	static protected $app_title;

	/**
	 * Stat something in the application directory
	 *
	 * An implementation can eg. call the appdir method of one level higher, to populate the stat cache
	 * with everything allowed on that level --> StreamWrapper::url_stat then looks up the cache
	 *
	 * @param string $id
	 */
	static public function stat($id)
	{
		//echo "<p>".__METHOD__."($id)</p>\n";

		// we just call appdir to preseed the cache with everything from one level higher
		$pattern = StreamWrapper::virtual_parent($id);

		static $appdir_called_for=null;
		if ($pattern !== $appdir_called_for)
		{
			$appdir_called_for = $pattern;
			self::appdir($pattern);
		}
	}

	/**
	 * Return directory entries for addressbook and populate the stat cache
	 *
	 * - root contains folder for each tracker queue
	 * - next level contains open tickets (or hashes starting with lang('open')) and dirs for status pending and closed (evtl with hashes)
	 *
	 * @param string $id
	 * @return array
	 */
	static public function appdir($id)
	{
		if ($id)
		{
			list($label,$hash) = explode(StreamWrapper::VIRTUAL_DELIMITER,$id);
			if ($label != 'all' && $label != lang('all') && !($tracker = array_search($label,self::$tracker->trackers)))
			{
				// check if some chars in the queue-name have been sanitized
				foreach(self::$tracker->trackers as $t => $name)
				{
					if ($label == StreamWrapper::sanitize_filename($name))
					{
						$tracker = $t;
						break;
					}
				}
				if (!$tracker) return null;
			}
			//echo '<p>'.__METHOD__."('$id') --> label=$label, hash=$hash --> filter=$filter</p>\n";
			if (!$hash)
			{
				$hashes = self::get_hashes($tracker);
				if ($hashes['all'] <= StreamWrapper::MAX_ENTRIES_NO_HASH)
				{
					$entries = self::entries($label,null,$tracker,'open');	// show entries without hashes
				}
				else
				{
					unset($hashes['all']);	// our all hash is lang($status)
					$entries = StreamWrapper::hashes2cache($hashes,self::APP_NAME,
						$label.StreamWrapper::VIRTUAL_DELIMITER);
				}
				foreach(array('pending','closed','not-closed') as $status)
				{
					$hashes = self::get_hashes($tracker,$status);
					unset($hashes['all']);	// our all hash is lang($status)
					// only one entry (plus all) --> remove it and show only all entry
					if (count($hashes) == 2)
					{
						array_shift($hashes);
					}
					$entries = array_merge($entries,StreamWrapper::hashes2cache($hashes,self::APP_NAME,
						$label.StreamWrapper::VIRTUAL_DELIMITER));
				}
			}
			else
			{
				$entries = self::entries($label,$hash,$tracker,'open');
			}
		}
		else	// --> add queues
		{
			foreach(array('all')+self::$tracker->trackers as $name)
			{
				$name = StreamWrapper::sanitize_filename($name);	//remove or replace not allowed chars
				$label = $name == 'all' ? lang('all') : $name;	// queues are not translated
				$link = $name.StreamWrapper::VIRTUAL_DELIMITER;
				StreamWrapper::addStatCacheDir(self::APP_NAME.'/'.$link,'egw/'.self::APP_NAME);
				if ($name != $label)
				{
					StreamWrapper::addStatCacheLink(self::APP_NAME.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,$link);
				}
				if (self::$app_title != self::APP_NAME)
				{
					StreamWrapper::addStatCacheLink(self::$app_title.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,'../'.self::APP_NAME.'/'.$link);

					if ($name != $label) StreamWrapper::addStatCacheLink(self::$app_title.'/'.$link,'../'.self::APP_NAME.'/'.$link);
				}
				$entries[] = $label.StreamWrapper::VIRTUAL_DELIMITER;
			}
		}
		//_debug_array($entries);
		if (StreamWrapper::DEBUG > 1) error_log(__METHOD__."('$id') = ".array2string($entries));
		return $entries;
	}

	/**
	 * Get the main projects with a given state and evtl. year
	 *
	 * @param string $label queue name
	 * @param string $hash year or year-month or 'all'
	 * @param int $tracker tr_tracker or NULL for all
	 * @param string $status ='open'
	 * @return array
	 */
	static private function entries($label,$hash,$tracker,$status='open')
	{
		if ($hash)
		{
			list($status,$year,$month_start) = $parts = explode('-',$hash);
			if ($year && !is_numeric($year))
			{
				unset($year); unset($month_start);
				list($status,$status2,$year,$month_start) = $parts;
				$status .= '-'.$status2;
			}
			switch($status)
			{
				case lang('open'):    $status = 'open'; break;
				case lang('closed'):  $status = 'closed'; break;
				case lang('pending'): $status = 'pending'; break;
				case lang('not-closed'): $status = 'not-closed'; break;
			}
		}
		$filter = array(
			'tr_status' => $status,
		);
		if ($tracker) $filter['tr_tracker'] = $tracker;

		if($year)
		{
			if (!$month_start)
			{
				$month_start = 1;
				$month_end   = 12;
			}
			else
			{
				$month_end = $month_start;
			}
			$filter[] = mktime(0,0,0,$month_start,1,$year).' <= '.self::$hash_column;
			$filter[] = self::$hash_column.' < '.mktime(0,0,0,$month_end+1,1,$year);
		}
		//echo "<p>".__METHOD__."($label,$hash,$tracker,$status) filter=".array2string($filter)."</p>\n";
		$entries = array();
		foreach((array)self::$tracker->search('',array('tr_id',self::$hash_column,'tr_summary','tr_tracker'),
			self::$hash_column,'','',false,'AND',false,$filter) as $ticket)
		{
			$id = $ticket['tr_id'];
			//$title = ($ticket[self::$hash_column] ? date('Y-m-d',$ticket[self::$hash_column]).' ' : '').self::$tracker->link_title($ticket);
			$title = ($ticket[self::$hash_column] ? date('Y-m-d',$ticket[self::$hash_column]).' ' : '').self::$tracker->trackers[$ticket['tr_tracker']].' - '.$ticket['tr_summary'];
			$entries[] = $name = StreamWrapper::entry2name(self::APP_NAME,$id,$title);
			StreamWrapper::entry2cache(self::APP_NAME,$id,$name,null,$label.$hash);
		}
		//_debug_array($entries);
		return $entries;
	}

	/**
	 * Get the hash values (eg. creation years) of existing projects of the given status
	 *
	 * @param int $tracker
	 * @param string $status ='open'
	 * @return array
	 */
	static private function get_hashes($tracker,$status='open')
	{
		static $cache=array();
		$hashes =& $cache[$tracker.':'.$status.':'.self::$hash_format];

		if(!isset($hashes))
		{
			//echo "<p>".__METHOD__."($tracker,$status) hash_format='".self::$hash_format."' ".function_backtrace()."</p>\n";
			$filter = array(
				'tr_status' => $status,
			);
			if ($tracker) $filter['tr_tracker'] = $tracker;

			$hashes = array();
			foreach((array)self::$tracker->search('',array(
					$GLOBALS['egw']->db->date_format($GLOBALS['egw']->db->from_unixtime(self::$hash_column),self::$hash_format).' AS hash',
					'COUNT(*) AS num',
				),'GROUP BY hash ORDER BY hash DESC','','',false,'AND',false,$filter,false) as $row)
			{
				// if more then MAX_ENTRIES_ALL=200 entries in one hash dir, automatic seitch to YEAR-MONTH hashes
				if (self::$hash_format == self::HASH_YEAR && $row['num'] > StreamWrapper::MAX_ENTRIES_ALL)
				{
					self::$hash_format = self::HASH_YEAR_MONTH;
					return self::get_hashes($tracker,$status);
				}
				$hashes[lang($status).'-'.$row['hash']] = $row['num'];
				$total += $row['num'];
			}
			$hashes['all'] = (int)$total;
			if ($total <= StreamWrapper::MAX_ENTRIES_ALL)
			{
				$hashes[lang($status)] = (int)$total;

				// adding symlink for the "all" hash named here by the status
				StreamWrapper::addStatCacheLink(self::$app_title.'/'.self::$tracker->trackers[$tracker].
					StreamWrapper::VIRTUAL_DELIMITER.'/'.lang($status).StreamWrapper::VIRTUAL_DELIMITER,
					'../'.self::$tracker->trackers[$tracker].StreamWrapper::VIRTUAL_DELIMITER.lang($status).StreamWrapper::VIRTUAL_DELIMITER);
			}
			//_debug_array($hashes);
		}
		return $hashes;
	}

	/**
	 * Init our static properties
	 *
	 */
	static function init_static()
	{
		self::$tracker = new tracker_bo();

		// add addressbook translations
		translation::add_app(self::APP_NAME);

		self::$app_title = $GLOBALS['egw_info']['user']['apps'][self::APP_NAME]['title'];
	}
}

Tracker::init_static();
