<?php
/**
 * Stylite: VFS - stream wrapper to merge (readonly) physical filesystem with sqlfs to make it writable
 *
 * @link http://www.stylite.de
 * @package stylite
 * @subpackage vfs
 * @author Ralf Becker <rb-AT-stylite.de>
 * @copyright (c) 2011-15 by Ralf Becker <RalfBecker-AT-stylite.de>
 * @version $Id: StreamWrapper.php 3556 2015-01-26 09:22:49Z ralfbecker $
 */

namespace EGroupware\Stylite\Vfs\Merge;

use EGroupware\Api\Vfs;

/**
 * stylite.merge - stream wrapper to merge (readonly) physical filesystem with sqlfs to make it writable
 *
 * Used mount print templates updated with EGroupware sources modifiable into the VFS.
 *
 * Examples:
 * mount stylite.merge://default/templates?merge=stylite/templates&lang=1&level=1 /templates
 * mount stylite.merge://default/etemplates?merge=.&lang=0&level=1&extension=xet&url=egw /etemplates
 *
 * Merges $level level directory sqlfs://default/templates over EGW_SERVER_ROOT /stylite/templates
 * - merge=path gives an absolute or EGw relative path, over which sqlfs is mounted
 * - level=N is necessry as streamwrapper does not know to what level a given path is part of the mountpoint
 * - lang=1 use user-language specific sub-directories, if they exist
 * - extension=xet only show original files with given extension
 *   for extension=xet removing a template will restore it's original state (deleting original template gives an error)
 * - url=egw specify download-url or "egw" to use for merged files from regular filesystem
 *
 * All write operations are direkted to sqlfs.
 * If a file get deleted an empty file is created as an marker for that in sqlfs.
 *
 * @todo support sqlfs operation mode "storage=db", currenty we always use the default storage=fs!
 * @todo fully support partial writing, currently file opened for writing on filesystem (not already in sqlfs) is empty!
 * @link http://www.php.net/manual/en/function.stream-wrapper-register.php
 */
class StreamWrapper extends Vfs\Sqlfs\StreamWrapper
{
	/**
	 * Scheme / protocol used for this stream-wrapper
	 */
	const SCHEME = 'stylite.merge';

	/**
	 * How much should be logged to the apache error-log
	 *
	 * 0 = Nothing
	 * 1 = only errors
	 * 2 = all function calls and errors (contains passwords too!)
	 */
	const LOG_LEVEL = 1;

	/**
	 * Language of the user, eg. "en", "de", "pt-br", ...
	 *
	 * @var string
	 */
	private static $user_lang = 'en';

	/**
	 * This method is called immediately after your stream object is created.
	 *
	 * @param string $url URL that was passed to fopen() and that this object is expected to retrieve
	 * @param string $mode mode used to open the file, as detailed for fopen()
	 * @param int $options additional flags set by the streams API (or'ed together):
	 * - STREAM_USE_PATH      If path is relative, search for the resource using the include_path.
	 * - STREAM_REPORT_ERRORS If this flag is set, you are responsible for raising errors using trigger_error() during opening of the stream.
	 *                        If this flag is not set, you should not raise any errors.
	 * @param string $opened_path full path of the file/resource, if the open was successfull and STREAM_USE_PATH was set
	 * @return boolean true if the ressource was opened successful, otherwise false
	 */
	function stream_open ( $url, $mode, $options, &$opened_path )
	{
		$this->opened_stream = null;

		$stat = self::url_stat($url, STREAM_URL_STAT_QUIET);

		// handle open for writing of a file in readonly merged filesystem --> pass it to parent
		if (in_array(substr(str_replace('b','',$mode),0,1), array('w', 'x', 'a', 'c')) || $mode == 'r+' ||
			$stat && strpos($stat['src_url'], parent::SCHEME) === 0)
		{
			$path = Vfs::parse_url($url, PHP_URL_PATH);
			if (!parent::url_stat(Vfs::dirname($path), 0))
			{
				self::mkdir(Vfs::dirname($url), 075, STREAM_MKDIR_RECURSIVE);
			}
			// if we have a previously removed file (0 byte file in parent), we need to remove that empty file first,
			// as sqlfs now calls static::url_stat, which tells it is not there and leaves us with 2 active files!
			if (!$stat && parent::url_stat($url, STREAM_URL_STAT_QUIET))
			{
				parent::unlink($url);
			}
			parent::stream_open($path, $mode, $options, $opened_path);	// sets $this->opened_stream and $this->opened_fs_id
		}
		elseif ($stat)
		{
			$this->opened_stream = fopen($stat['src_url'], $mode);
			$this->opened_path = $url;
		}
		if (self::LOG_LEVEL > 1 || self::LOG_LEVEL && !is_resource($this->opened_stream))
		{
			error_log(__METHOD__."('$url', '$mode', $options) fopen('$stat[src_url]', '$mode')=".array2string($this->opened_stream).' returning '.array2string(is_resource($this->opened_stream)));
		}
		return is_resource($this->opened_stream);
	}

	/**
	 * This method is called when the stream is closed, using fclose().
	 *
	 * You must release any resources that were locked or allocated by the stream.
	 */
	function stream_close ( )
	{
		if ($this->opened_fs_id)
		{
			$ret = parent::stream_close();
		}
		else
		{
			$ret = fclose($this->opened_stream);
			$this->opened_stream = $this->opened_path = null;
		}
		return $ret;
	}

	/**
	 * This method is called in response to unlink() calls on URL paths associated with the wrapper.
	 *
	 * It should attempt to delete the item specified by path.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support unlinking!
	 *
	 * @param string $url
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function unlink ( $url )
	{
		// for extension=xet, remove restores original file or fails, if one tries to delete the orginal file
		$merge = $lang = $level = $extension = null;
		self::parse_url_query($url, $merge, $lang, $level, $extension);
		// if file exists our in overlay, remove it from there (evtl. restoring original file!)
		if (parent::url_stat($url, STREAM_URL_STAT_QUIET))
		{
			return parent::unlink($url);
		}
		// never allow to delete original xet file, as it renders egw unusable
		elseif ($extension === 'xet')
		{
			return false;
		}
		// to "delete" a file, create an empty (size=0) file (in sqlfs)
		if (($f = fopen($url, 'w')))
		{
			fclose($f);
			return true;
		}
		return false;
	}

	/**
	 * This method is called in response to rename() calls on URL paths associated with the wrapper.
	 *
	 * It should attempt to rename the item specified by path_from to the specification given by path_to.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support renaming.
	 *
	 * The regular filesystem stream-wrapper returns an error, if $url_from and $url_to are not either both files or both dirs!
	 *
	 * @param string $url
	 * @param string $to
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function rename ( $url, $to )
	{
		if (!($stat = self::url_stat($url, 0)))
		{
			return false;
		}
		// for files already in sqlfs, let parent handle it
		if (strpos($stat['src_url'], parent::SCHEME) === 0)
		{
			return parent::rename($url, $to);
		}
		// otherwise copy source of merge and unlink it then (empty file)
		return copy($stat['src_url'], $to) && self::unlink($url);
	}

	/**
	 * This method is called in response to mkdir() calls on URL paths associated with the wrapper.
	 *
	 * It should attempt to create the directory specified by path.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support creating directories.
	 *
	 * @param string $url url or path, url is needed to create parents with correct rights
	 * @param int $mode
	 * @param int $options Posible values include STREAM_REPORT_ERRORS and STREAM_MKDIR_RECURSIVE
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function mkdir ( $url, $mode, $options )
	{
		// check if mountpoint exits and if not create it with 075 and group Admins
		$merge = $lang = $level = null;
		if ($url[0] != '/' && self::parse_url_query($url, $merge, $lang, $level))
		{
			$mountpoint = implode('/', array_slice(explode('/', Vfs::parse_url($url, PHP_URL_PATH)), 0, 1+$level));
			if (!parent::url_stat($mountpoint, STREAM_URL_STAT_QUIET))
			{
				Vfs::$is_root = true;
				if (parent::mkdir($mountpoint, 075, 0))
				{
					Vfs::chmod($mountpoint, 075);
					Vfs::chown($mountpoint, 0);
					Vfs::chgrp($mountpoint, abs($GLOBALS['egw']->accounts->name2id('Admins', 'account_lid', 'g')));
				}
				Vfs::$is_root = false;
			}
		}
		// not all parents might exist, because of merge: setting STREAM_MKDIR_RECURSIVE
		return parent::mkdir( $url, $mode, $options|STREAM_MKDIR_RECURSIVE );
	}

	/**
	 * parse a query containing mount parameters: merge
	 *
	 * @param string|array $url url string or array returned by parse_url
	 * @param string &$merge path to merge with current directory
	 * @param boolean &$lang=false should we use a lang-specific path
	 * @param int &$level=1 hiearchy level of the mountpoint, eg. 1 for /templates or 2 for /home/someone
	 * @param string &$extension=null optional extension parameter
	 * @param string &$durl=null optional (download) url parameter
	 * @return boolean true on successfull parse, false on error
	 */
	private static function parse_url_query($url, &$merge, &$lang=false, &$level=1, &$extension=null, &$durl=null)
	{
		$lang = false;
		$level = 1;
		$durl = $extension = null;
		$url_parts = is_array($url) ? $url : Vfs::parse_url($url);
		$params = null;
		parse_str($url_parts['query'], $params);

		foreach($params as $name => $value)
		{
			switch($name)
			{
				case 'merge':
					$merge = $value;
					if (!($merge[0] == '/' || strtoupper(substr(PHP_OS, 0, 3)) === 'WIN' && ($merge[0] == '\\' || $merge[1] == ':')))
					{
						$merge = EGW_SERVER_ROOT.'/'.$merge;
					}
					if (!file_exists($merge) || !is_dir($merge))
					{
						if (self::LOG_LEVEL) error_log(__METHOD__.'('.array2string($url).") merge path '$merge' is NO directory!");
						return false;
					}
					break;
				case 'lang':
					$lang = (boolean)$value;
					break;
				case 'level':
					$level = (int)$value;
					break;
				case 'extension':
					$extension = $value;
					break;
				case 'url':
					$durl = $value;
					break;
				default:
					if (self::LOG_LEVEL) error_log(__METHOD__.'('.array2string($url).") unknown parameter '$name'!");
					return false;
			}
		}
		// adapt $merge to given path in $url
		$path = $url_parts['path'];
		if ($level)
		{
			$path_parts = explode('/', $path);
			array_shift($path_parts);	// remove empty part before 1. slash
			for ($i = 0; $i < $level; ++$i)
			{
				array_shift($path_parts);
			}
			array_unshift($path_parts, '');
			$path = implode('/', $path_parts);
		}
		if ($path != '/') $merge .= $path;

		if (self::LOG_LEVEL > 1) error_log(__METHOD__.'('.array2string($url).") returning merge='$merge' and lang=$lang");
		return true;
	}

	private $dir_entries;

	/**
	 * This method is called immediately when your stream object is created for examining directory contents with opendir().
	 *
	 * @param string $path URL that was passed to opendir() and that this object is expected to explore.
	 * @return booelan
	 */
	function dir_opendir ( $url, $options )
	{
		$url_parts = Vfs::parse_url($url);
		$path = $url_parts['path'];
		$merge = $use_lang = $level = $extension = null;
		if (!self::parse_url_query($url_parts, $merge, $use_lang, $level, $extension))
		{
			return false;
		}
		if (parent::url_stat($path, 0) && parent::dir_opendir($path, $options))
		{
			$this->dir_entries = array();
			while(($file = parent::dir_readdir()))
			{
				$this->dir_entries[] = $file;
			}
			parent::dir_closedir();
		}
		if (file_exists($merge) && is_dir($merge))
		{
			if (!is_array($this->dir_entries)) $this->dir_entries = array();
			if ($use_lang && (file_exists($lang_merge=$merge.'/'.self::$user_lang) || file_exists($lang_merge=$merge.'/en')))
			{
				$merged = scandir($lang_merge);
			}
			else
			{
				$merged = scandir($merge);
			}
			foreach($merged as $file)
			{
				// ignore ".", ".." and not readable files
				if ($file == '.' || $file == '..' || !(is_readable($merge.'/'.$file) || is_readable($lang_merge.'/'.$file)))
				{
					//error_log(__METHOD__.'('.array2string($url).") removed '$file' because NOT readable or . or ..");
					continue;
				}
				// remove files not matching given extension
				if ($extension && !is_dir($merge.'/'.$file) && substr($file, -strlen($extension)-1) !== '.'.$extension)
				{
					//error_log(__METHOD__."() removed $file because not having extension $extension!");
					continue;
				}
				if (!in_array($file, $this->dir_entries))
				{
					$this->dir_entries[] = $file;
				}
				// empty files are considered deleted, to give admin a chance to "remove" merged files from his users
				elseif (($stat = parent::url_stat(Vfs::concat($path, $file), 0)) &&
					$stat['mode'] & self::MODE_FILE && !$stat['size'])
				{
					if (self::LOG_LEVEL > 1) error_log(__METHOD__.'('.array2string($url).") removed '$file' because size=0");
					unset($this->dir_entries[array_search($file, $this->dir_entries)]);
				}
			}
		}
		if (!is_array($this->dir_entries))
		{
			if (self::LOG_LEVEL) error_log(__METHOD__.'('.array2string($url).") returning FALSE (url no directory!)");
			return false;
		}
		if ($this->dir_entries) sort($this->dir_entries);
		if (self::LOG_LEVEL > 1) error_log(__METHOD__.'('.array2string($url).") contains ".array2string($this->dir_entries).' returning TRUE');

		reset($this->dir_entries);	// position array pointer on first element!

		return true;
	}

	/**
	 * This method is called in response to stat() calls on the URL paths associated with the wrapper.
	 *
	 * It should return as many elements in common with the system function as possible.
	 * Unknown or unavailable values should be set to a rational value (usually 0).
	 *
	 * If you plan to use your wrapper in a require_once you need to define stream_stat().
	 * If you plan to allow any other tests like is_file()/is_dir(), you have to define url_stat().
	 * stream_stat() must define the size of the file, or it will never be included.
	 * url_stat() must define mode, or is_file()/is_dir()/is_executable(), and any of those functions affected by clearstatcache() simply won't work.
	 * It's not documented, but directories must be a mode like 040777 (octal), and files a mode like 0100666.
	 * If you wish the file to be executable, use 7s instead of 6s.
	 * The last 3 digits are exactly the same thing as what you pass to chmod.
	 * 040000 defines a directory, and 0100000 defines a file.
	 *
	 * @param string $url
	 * @param int $flags holds additional flags set by the streams API. It can hold one or more of the following values OR'd together:
	 * - STREAM_URL_STAT_LINK	For resources with the ability to link to other resource (such as an HTTP Location: forward,
	 *                          or a filesystem symlink). This flag specified that only information about the link itself should be returned,
	 *                          not the resource pointed to by the link.
	 *                          This flag is set in response to calls to lstat(), is_link(), or filetype().
	 * - STREAM_URL_STAT_QUIET	If this flag is set, your wrapper should not raise any errors. If this flag is not set,
	 *                          you are responsible for reporting errors using the trigger_error() function during stating of the path.
	 *                          stat triggers it's own warning anyway, so it makes no sense to trigger one by our stream-wrapper!
	 * @return array
	 */
	static function url_stat ( $url, $flags )
	{
		$stat = parent::url_stat($url, $flags|STREAM_URL_STAT_QUIET);

		$merge = $use_lang = $level = $extension = null;
		if ($stat)
		{
			// check if file got deleted (zero size)
			if ($stat['mode'] & self::MODE_FILE && !$stat['size'])
			{
				$stat = false;
			}
			else
			{
				$stat['src_url'] = parent::SCHEME.'://default'.Vfs::parse_url($url, PHP_URL_PATH);
			}
		}
		// file not in sqlfs, but maybe in merged directories
		elseif (!$stat && self::parse_url_query($url, $merge, $use_lang, $level, $extension))
		{
			if ($use_lang && !is_dir($merge))
			{
				$parent_dir = Vfs::dirname($merge);
				$lang_dir = file_exists($parent_dir.'/'.self::$user_lang) ? $parent_dir.'/'.self::$user_lang :
					(file_exists($parent_dir.'/en') ? $parent_dir.'/en' : false);
				$lang_merge = $lang_dir.'/'.basename($merge);
				if ($lang_dir && ($stat = $flags & STREAM_URL_STAT_LINK ? @lstat($lang_merge) : @stat($lang_merge)))	// suppressed the stat failed warnings
				{
					$merge = $lang_merge;
				}
			}
			// fail if extension given and not matched
			if ($extension &&  !is_dir($merge) && substr($merge, -strlen($extension)-1) !== '.'.$extension)
			{
				$stat = false;
			}
			elseif (!$stat)
			{
				$stat = $flags & STREAM_URL_STAT_LINK ? @lstat($merge) : @stat($merge);	// suppressed the stat failed warnings
			}
			if ($stat)
			{
				$stat['src_url'] = $merge;
				// setting mode from first real sqlfs parent
				$parent_dir = Vfs::dirname(Vfs::parse_url($url, PHP_URL_PATH));
				while ($parent_dir != '/' && !($parent_stat = parent::url_stat($parent_dir, 0)))
				{
					$parent_dir = Vfs::dirname($parent_dir);
				}
				if (!$parent_stat) $parent_stat = array(
					'uid' => 0,
					'gid' => abs($GLOBALS['egw']->accounts->name2id('Admins', 'account_lid', 'g')),
					'mode' => self::MODE_DIR | 075,
				);
				$stat['uid'] = $stat[4] = $parent_stat['uid'];
				$stat['gid'] = $stat[5] = $parent_stat['gid'];
				$stat['mode'] = $stat[2] = $stat['mode'] & self::MODE_DIR ? $parent_stat['mode'] : self::MODE_FILE|($parent_stat['mode']&0666);
				$stat['ino'] = -$stat['ino'];	// make non-sqlfs ino's negative for propfind to detect
			}
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__.'('.array2string($url).", $flags) returning ".array2string($stat).' '.function_backtrace());
		return $stat;
	}

	/**
	 * This method is called in response to readdir().
	 *
	 * It should return a string representing the next filename in the location opened by dir_opendir().
	 *
	 * @return string
	 */
	function dir_readdir ( )
	{
		if (($ret = $this->dir_entries ? current($this->dir_entries) : false))
		{
			next($this->dir_entries);
		}
		return $ret;
	}

	/**
	 * This method is called in response to rewinddir().
	 *
	 * It should reset the output generated by dir_readdir(). i.e.:
	 * The next call to dir_readdir() should return the first entry in the location returned by dir_opendir().
	 *
	 * @return boolean
	 */
	function dir_rewinddir ( )
	{
		if (is_array($this->dir_entries))
		{
			reset($this->dir_entries);
			return true;
		}
		return false;
	}

	/**
	 * Read properties for a ressource (file, dir or all files of a dir)
	 *
	 * Reimplemented to quiten error_log if file is not in sqlfs
	 *
	 * @param array|string|int $path_ids (array of) string with path or integer fs_id
	 * @param string $ns ='http://egroupware.org/' namespace if propfind should be limited to a single one, use null for all
	 * @return array|boolean false on error ($path_ids does not exist), array with props (values for keys 'name', 'ns', 'value'), or
	 * 	fs_id/path => array of props for $depth==1 or is_array($path_ids)
	 */
	static function propfind($path_ids,$ns=Vfs::DEFAULT_PROP_NAMESPACE)
	{
		if (($is_string = !is_array($path_ids))) $path_ids = array($path_ids);

		foreach($path_ids as $n => &$url)
		{
			if (!is_numeric($url) && ($stat = self::url_stat($url, 0)))
			{
				$url = $stat['ino'];
			}
			if ((int)$url <= 0) unset($path_ids[$n]);	// no fs_id from sqlfs --> ignore, as it only gives an error_log
		}
		if ($path_ids)
		{
			return parent::propfind($is_string ? $path_ids[0] : $path_ids, $ns);
		}
		return array();
	}

	/**
	 * This method is called in response to closedir().
	 *
	 * You should release any resources which were locked or allocated during the opening and use of the directory stream.
	 *
	 * @return boolean
	 */
	function dir_closedir ( )
	{
		$this->opened_dir = $this->dir_entries = null;

		return true;
	}

	/**
	 * URL to download a file
	 *
	 * We use our webdav handler as download url instead of an own download method.
	 * The webdav hander (filemanager/webdav.php) recognices eGW's session cookie and of cause understands regular GET requests.
	 *
	 * @param string $_url
	 * @param boolean $force_download =false add header('Content-disposition: filename="' . basename($path) . '"'), currently not supported!
	 * @todo get $force_download working through webdav
	 * @return string|false string with full download url or false to use default webdav.php url
	 */
	static function download_url($_url,$force_download=false)
	{
		unset($force_download);	// not used, but required by function signature
		// no download url for sqlfs / merged content
		if (!($stat = self::url_stat($_url, 0)) || strpos($stat['src_url'], parent::SCHEME) === 0)
		{
			return false;
		}
		list(,$query) = explode('?',$_url,2);
		$get = null;
		parse_str($query, $get);
		if (empty($get['url'])) return false;	// no download url given for this mount-point

		if ($get['url'] === 'egw') $get['url'] = $GLOBALS['egw_info']['server']['webserver_url'];

		if (!($mount_url = Vfs::mount_url($_url))) return false;	// no mount url found, should not happen
		list($mount_url) = explode('?',$mount_url);

		list($url) = explode('?', $_url);
		$relpath = substr($url,strlen($mount_url));

		$download_url = Vfs::concat($get['url'],$relpath);
		if ($download_url[0] == '/')
		{
			$download_url = ($_SERVER['HTTPS'] ? 'https://' : 'http://').
				$_SERVER['HTTP_HOST'].$download_url;
		}

		//die(__METHOD__."('$url') --> relpath = $relpath --> $download_url");
		return $download_url;
	}

	/**
	 * Init our static properties and register this wrapper
	 *
	 */
	static function init_static()
	{
		stream_wrapper_register(self::SCHEME, __CLASS__);

		if (!(self::$user_lang = $GLOBALS['egw_info']['user']['preferences']['common']['lang'])) self::$user_lang = 'en';
	}
}

StreamWrapper::init_static();
