<?php
/**
 * Stylite: VFS - stream wrapper with versioning
 *
 * @link http://www.stylite.de
 * @package stylite
 * @subpackage vfs
 * @author Ralf Becker <rb-AT-stylite.de>
 * @copyright (c) 2010-15 by Ralf Becker <RalfBecker-AT-stylite.de>
 * @version $Id: StreamWrapper.php 3962 2015-08-15 09:37:30Z ralfbecker $
 */

namespace EGroupware\Stylite\Vfs\Versioning;

use EGroupware\Api\Vfs;

// explicitly import old phpgwapi classes used:
use egw_db;
use egw_framework;
use egw_exception_assertion_failed;
use mime_magic;

/**
 * Stylite: stream wrapper with versioning
 *
 * The versioning stream wrapper has one optional mount option: min_version=N, with N is an integer
 * number of seconds required to create a new version, otherwise the current one get updated:
 *
 * filemanager/cli.php mount --user root_admin --passwd secret \
 * 	'stylite.versioning://default/versioned?min_version=600' /versioned
 *
 * The above example remounts the (existing!) directory /versioned with versioning.
 * Creating new a version only if the current one is older then 10min.
 *
 * The versioning works on top of the regular (open source) Vfs\Sqlfs\StreamWrapper.
 * Only the most up to date version has fs_active=true set, all (older) versions
 * and also deleted ones have fs_active=false set AND get ignored by Vfs\Sqlfs\StreamWrapper.
 *
 * A new version is created with fs_active=false and:
 * a) empty for mode: 'w', 'w+'
 * b) with content of previous version for mode: 'r+', 'a', 'a+'
 * In case of b) $mode[0] != 'w', we only create a new version if write was called!
 * After closeing the file the fs_active flag is swaped:
 * UPDATE egw_sqlfs SET fs_active=CASE fs_id WHEN $new_id THEN true ELSE false WHERE fs_name=$name and fs_dir=$dir_id
 *
 * Deleting a file or directory also sets fs_active=false (do we want this for dirs with empty attic too?),
 *
 * @ToDo Done/Check: stream_open() copy old file for modes 'r+', 'a' and 'a+' as they do NOT trunc the file
 * @ToDo Done/Check: stream_open() copy metadata (uid, gid, mode, created(?) and props from old version)
 * @ToDo Done/Check: unlink(), rmdir()
 * @ToDo Done/Check: rename() do we rename the versions too: yes, implemented in sqlfs
 * @ToDo Done/Check: dir_*() to display versions of a file and the attic (deleted files) of a dir
 * 	by prepending to a file '.versions-'.$filename or appending to a dir $dir.'/.versions/' or $dir.'/.attic/'
 * @ToDo Done/Check: add mount parameter for minimum version age (no new version if current one is less then x old)
 *  easy to do in stream_open, just pass call to sqlfs in that case, what is a reasonable default 1h or 0s?
 * @ToDo Done/Check: re-activate deleted dirs by using mkdir() with same name (.attic still contains old deleted stuff)
 * @ToDo Done/Check: a new version should inherit the props
 *
 * @ToDo Done/Check: get versioning working with links_stream_wrapper or stylite_links_stream_wrapper (which extend sqlfs)
 * @ToDo Done/Check: deleting of files in attic or old versions depending on server config allow_delete_versions
 * @ToDo Administrativ tool to eg. delete versions older then x days or closer then y hours or even all versions
 *
 * The stream wrapper interface is according to the docu on php.net
 * @link http://www.php.net/manual/en/function.stream-wrapper-register.php
 */
class StreamWrapper extends Vfs\Sqlfs\StreamWrapper
{
	/**
	 * Scheme / protocol used for this stream-wrapper
	 */
	const SCHEME = 'stylite.versioning';
	/**
	 * Prefix to predend to get an url from a path
	 */
	const PREFIX = 'stylite.versioning://default';

	/**
	 * Name of versions directory without leading dot
	 */
	const VERSIONS = 'versions';
	/**
	 * Name of attic directory without leading dot
	 */
	const ATTIC = 'attic';

	/**
	 * Filename PREG for MS Office for files renamed before getting deleted in order to replace them
	 *
	 * Word docx: WRL\d{4}.tmp
	 * Excel xlsx: [0-9A-F]{8}.tmp
	 * PowerPoint pptx: [0-9A-F]{8}.tmp
	 * --> check just for .tmp
	 */
	const OFFICE_TMP_RENAME_DELETE = '/\.tmp$/i';
	/**
	 * Name of attribute used to store original name, when office renames file in order to replace it
	 */
	const OFFICE_ORIGINAL_NAME_ATTR = 'office-orginal-name';

	/**
	 * Temporary files for Office 2010, which make no sense to be versioned
	 *
	 * all starting with ~$ (tilde dolar) and ending in orignal filename, eg. ~$test.docx
	 */
	const OFFICE_TMP_NO_VERSIONING = '/^~\$/';

	/**
	 * How much should be logged to the apache error-log
	 *
	 * 0 = Nothing
	 * 1 = only errors
	 * 2 = all function calls and errors (contains passwords too!)
	 * 3 = log line numbers in sql statements
	 */
	const LOG_LEVEL = 1;

	/**
	 * Minimum time between versions
	 *
	 * A new version get only created if the current version is older then min_version_time seconds
	 *
	 * @var int
	 */
	public $min_version=0;

	/**
	 * Opened file got created as new version with fs_active=0
	 *
	 * @var boolean
	 */
	protected $new_version;

	/**
	 * How oven was write called, to track if we need to create a new version
	 *
	 * @var int
	 */
	protected $write_called=0;

	/**
	 * Props of old version of opened file, to add them if file got closed
	 *
	 * @var array
	 */
	protected $props_opened_file;

	/**
	 * This method is called immediately after your stream object is created.
	 *
	 * Reimplemented from sqlfs to create a new version instead opening an existing one.
	 *
	 * @param string $url URL that was passed to fopen() and that this object is expected to retrieve
	 * @param string $_mode mode used to open the file, as detailed for fopen()
	 * @param int $options additional flags set by the streams API (or'ed together):
	 * - STREAM_USE_PATH      If path is relative, search for the resource using the include_path.
	 * - STREAM_REPORT_ERRORS If this flag is set, you are responsible for raising errors using trigger_error() during opening of the stream.
	 *                        If this flag is not set, you should not raise any errors.
	 * @param string $opened_path full path of the file/resource, if the open was successfull and STREAM_USE_PATH was set
	 * @return boolean true if the ressource was opened successful, otherwise false
	 */
	function stream_open ( $url, $_mode, $options, &$opened_path )
	{
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url, $_mode, $options)");

		$path = Vfs::parse_url($url,PHP_URL_PATH);
		$mode = str_replace('b', '', $_mode);

		// check if we are mounted with a minimum time between versions
		if (($pos=strpos($url,'min_version=')) != false && ($url[$pos-1] == '?' || $url[$pos-1] == '&'))
		{
			$this->min_version = (int)substr($url,$pos+12);
			//error_log(__METHOD__."($url) using min_version_time=$this->min_version");
		}

		// opening for reading or not yet existing files are no concern to versioning --> sqlfs
		if ($mode == 'r' || $mode[0] == 'x' || !($stat = parent::url_stat($path,STREAM_URL_STAT_QUIET)) ||
			($stat['mode'] & self::MODE_FILE) && !$stat['size'] ||	// current version is empty --> no need to create new version
			$this->min_version > 0 && $stat['mtime'] > time()-$this->min_version ||
			$stat && ($stat['mode'] & self::MODE_DIR) == self::MODE_DIR)	// give "Is a directory" error
		{
			// we supply our own __CLASS__ for url_stat, to allow (readonly) opening of versions and attic files
			return parent::stream_open($url,$mode,$options,$opened_path,null,__CLASS__);
		}
		// get props of old version, to copy them to new one
		$this->props_opened_file = parent::propfind($path,null);

		// we do NOT need to check write access to parent, as parent::stream_open does that!
		$this->new_version = parent::stream_open($url,$mode,$options,$opened_path,array(
			'fs_active' => self::_pdo_boolean(false),	// new version is made active on closing
			'fs_uid' => $stat['uid'],
			'fs_gid' => $stat['gid'],
			'fs_mode' => $stat['mode'] & 0777,
			'fs_created' => self::_pdo_timestamp($stat['ctime']),
			//'fs_creator' => self::$stat_cache[$path]['fs_creator'],
			'fs_size' => $mode[0] != 'w' ? $stat['size'] : 0,
		));

		// copy current content into new version, if mode != w (w truncs the file)
		if ($this->new_version && $mode[0] != 'w')
		{
			if (($src = fopen($url,'r')))
			{
				if ($mode[0] == 'a')	// position stream pointer to start again
				{
					$this->stream_seek(0,SEEK_SET);
				}
				stream_copy_to_stream($src,$this->opened_stream);
				fclose($src);

				if ($mode == 'r+')	// position stream pointer to start again
				{
					$this->stream_seek(0,SEEK_SET);
				}
			}
		}
		return $this->new_version;
	}

	/**
	 * This method is called in response to fwrite() calls on the stream.
	 *
	 * Reimplemented from sqlfs to track if something was written, if not there's no need
	 * to create a new version.
	 *
	 * You should store data into the underlying storage used by your stream.
	 * If there is not enough room, try to store as many bytes as possible.
	 * You should return the number of bytes that were successfully stored in the stream, or 0 if none could be stored.
	 * You must also update the read/write position of the stream by the number of bytes that were successfully written.
	 *
	 * @param string $data
	 * @return integer
	 */
	function stream_write ( $data )
	{
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($data)");

		++$this->write_called;

		return parent::stream_write($data);
	}

	/**
	 * This method is called when the stream is closed, using fclose().
	 *
	 * You must release any resources that were locked or allocated by the stream.
	 */
	function stream_close ( )
	{
		if ($this->new_version)
		{
			if (!$this->write_called && $this->opened_mode[0] != 'w')
			{
				// no change in the file --> no need to create new version --> delete it
				$query = 'DELETE FROM '.self::TABLE.' WHERE fs_id=?';
				$params = array($this->opened_fs_id);
			}
			else
			{
				if (!($dir_stat = parent::url_stat(dirname($this->opened_path),STREAM_URL_STAT_QUIET)))
				{
					throw new egw_exception_assertion_failed(__METHOD__."() parent dir was removed while writing to $this->opened_path!");
				}
				// swap fs_active flag
				$query = 'UPDATE '.self::TABLE.' SET fs_active=CASE fs_id WHEN '.(int)$this->opened_fs_id.' THEN '.
					self::_pdo_boolean(true).' ELSE '.self::_pdo_boolean(false).' END WHERE fs_dir=? AND fs_name=?';
				$params = array(
					$dir_stat['ino'],
					basename($this->opened_path),
				);
				// copy props from old version
				if ($this->props_opened_file)
				{
					parent::proppatch($this->opened_fs_id,$this->props_opened_file);
				}
			}
		}
		$ok = parent::stream_close();

		if (isset($query))
		{
			$stmt = self::$pdo->prepare($query);
			$stmt->execute($params);
		}
		return $ok;
	}

	/**
	 * This method is called in response to unlink() calls on URL paths associated with the wrapper.
	 *
	 * Reimplemented to only set fs_active=false (instead deleteing the file).
	 *
	 * It should attempt to delete the item specified by path.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support unlinking!
	 *
	 * @param string $url
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function unlink ( $url )
	{
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url)");

		$path = Vfs::parse_url($url,PHP_URL_PATH);

		if (!($stat = self::url_stat($path,STREAM_URL_STAT_LINK)))
		{
			self::_remove_password($url);
			if (self::LOG_LEVEL) error_log(__METHOD__."($url) (stat=".array2string($stat).") no such file!");
			return false;	// no permission or file does not exist
		}
		if ($stat['mime'] == self::DIR_MIME_TYPE)
		{
			self::_remove_password($url);
			if (self::LOG_LEVEL) error_log(__METHOD__."($url) is NO file!");
			return false;	// no permission or file does not exist
		}
		if (!isset(self::$stat_cache[$path]))
		{
			throw new egw_exception_assertion_failed(__METHOD__."('$url') self::\$stat_cache['$path'] is NOT set!");
		}
		// check if already not active (file in attic or old version)
		if (isset(self::$stat_cache[$path]['fs_active']) && !egw_db::from_bool(self::$stat_cache[$path]['fs_active']))
		{
			if (!self::check_delete_version($url))
			{
				if (self::LOG_LEVEL) error_log(__METHOD__."($url) permission denied (final delete)!");
				return false;
			}
			// we can call parent::unlink(), as stat cache is now filled (calling it before would have failed!)
			return parent::unlink($url, self::url_stat(dirname(dirname($path)), STREAM_URL_STAT_LINK));
		}
		// check if this is an office temporary file, which makes no sense to version --> call real unlink from parent
		if (preg_match(self::OFFICE_TMP_NO_VERSIONING, basename($path)))
		{
			return parent::unlink($path);
		}
		if (!Vfs::check_access(dirname($path),Vfs::WRITABLE))
		{
			self::_remove_password($url);
			if (self::LOG_LEVEL) error_log(__METHOD__."($url) (stat=".array2string($stat).") permission denied!");
			return false;	// no permission
		}
		// check if file to delete is an office temp. file, which got renamed before deleted
		if (preg_match(self::OFFICE_TMP_RENAME_DELETE, $temp_name = basename($path)) &&
			($props = self::propfind($stat['ino'], Vfs::DEFAULT_PROP_NAMESPACE)))
		{
			foreach($props as $prop)
			{
				// check if we have an original name stored
				if ($prop['name'] == self::OFFICE_ORIGINAL_NAME_ATTR)
				{
					// --> remove stored property
					self::proppatch($stat['ino'], array(array(
						'name' => self::OFFICE_ORIGINAL_NAME_ATTR,
						'ns'   => Vfs::DEFAULT_PROP_NAMESPACE,
					)));
					// --> rename temp. file(s) / old version(s) back to original name, so they are visible in versioning
					$stmt = self::$pdo->prepare('UPDATE '.self::TABLE.' SET fs_name=:fs_name,fs_mime=:fs_mime WHERE fs_name=:temp_name');
					$ok = $stmt->execute($vals=array(
						'fs_name' => $prop['val'],
						'fs_mime' => mime_magic::filename2mime($prop['val'], true),
						'temp_name' => $temp_name,
					));
					error_log(__METHOD__."('$url') stmt=".$stmt->queryString."->execute(".array2string($vals).") returned ".array2string($ok));
					unset($stmt);
					break;
				}
			}
		}
		$stmt = self::$pdo->prepare('UPDATE '.self::TABLE.' SET fs_active='.self::_pdo_boolean(false).' WHERE fs_id=:fs_id');
		unset(self::$stat_cache[$path]);

		return $stmt->execute(array('fs_id' => $stat['ino']));
	}

	/**
	 * This method is called in response to rmdir() calls on URL paths associated with the wrapper.
	 *
	 * It should attempt to remove the directory specified by path.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support removing directories.
	 *
	 * Trying to finally remove a deleted directory still containing (deleted) files will fail with "dir is not empty!"
	 *
	 * @param string $url
	 * @param int $options Possible values include STREAM_REPORT_ERRORS.
	 * @return boolean TRUE on success or FALSE on failure.
	 */
	static function rmdir ( $url, $options )
	{
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url)");

		$path = Vfs::parse_url($url,PHP_URL_PATH);
		$parent = Vfs::dirname($path);

		if (!($stat = self::url_stat($path,0)) || $stat['mime'] != self::DIR_MIME_TYPE ||
			!Vfs::check_access($parent,Vfs::WRITABLE))
		{
			self::_remove_password($url);
			$err_msg = __METHOD__."($url,$options) ".(!$stat ? 'not found!' :
				($stat['mime'] != self::DIR_MIME_TYPE ? 'not a directory!' : 'permission denied!'));
			if (self::LOG_LEVEL) error_log($err_msg);
			if (!($options & STREAM_REPORT_ERRORS))
			{
				trigger_error($err_msg,E_USER_WARNING);
			}
			return false;	// no permission or file does not exist
		}
		$stmt = self::$pdo->prepare('SELECT COUNT(*) FROM '.self::TABLE.' WHERE fs_active=? AND fs_dir=?');
		$stmt->execute(array(self::_pdo_boolean(true),$stat['ino']));
		if ($stmt->fetchColumn())
		{
			self::_remove_password($url);
			if (self::LOG_LEVEL) error_log(__METHOD__."($url,$options) dir is not empty!");
			if (!($options & STREAM_REPORT_ERRORS))
			{
				trigger_error(__METHOD__."('$url',$options) dir is not empty!",E_USER_WARNING);
			}
			return false;
		}
		unset($stmt);	// free statement object, on some installs a new prepare fails otherwise!

		if (!isset(self::$stat_cache[$path]))
		{
			throw new egw_exception_assertion_failed(__METHOD__."('$url') self::\$stat_cache['$path'] is NOT set!");
		}
		// check if already not active (file in attic or old version)
		if (isset(self::$stat_cache[$path]['fs_active']) && !egw_db::from_bool(self::$stat_cache[$path]['fs_active']))
		{
			if (!self::check_delete_version($url))
			{
				if (self::LOG_LEVEL) error_log(__METHOD__."($url) permission denied (final delete)!");
				return false;
			}
			// we can call parent::unlink(), as stat cache is now filled (calling it before would have failed!)
			return parent::rmdir($url,$options);
		}
		$upd_stmt = self::$pdo->prepare('UPDATE '.self::TABLE.' SET fs_active='.self::_pdo_boolean(false).' WHERE fs_id=:fs_id');
		unset(self::$stat_cache[$path]);

		return $upd_stmt->execute(array('fs_id' => $stat['ino']));
	}

	/**
	 * This method is called in response to mkdir() calls on URL paths associated with the wrapper.
	 *
	 * Reimplemented to only activate a directory again, if it still exist in the attic
	 *
	 * It should attempt to create the directory specified by path.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support creating directories.
	 *
	 * @param string $url
	 * @param int $mode
	 * @param int $options Posible values include STREAM_REPORT_ERRORS and STREAM_MKDIR_RECURSIVE
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function mkdir ( $url, $mode, $options )
	{
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url,$mode,$options)");

		$path = Vfs::parse_url($url,PHP_URL_PATH);

		$attic_path = Vfs::dirname($path).'/.attic/'.Vfs::basename($path);
		if (($stat = self::url_stat($attic_path,STREAM_URL_STAT_QUIET)))
		{
			$stmt = self::$pdo->prepare('UPDATE '.self::TABLE.' SET fs_active='.self::_pdo_boolean(true).' WHERE fs_id=:fs_id');
			unset(self::$stat_cache[$path]);
			unset(self::$stat_cache[$attic_path]);

			return $stmt->execute(array('fs_id' => $stat['ino']));
		}
		return parent::mkdir($url,$mode,$options);
	}

	/**
	 * This method is called in response to stat() calls on the URL paths associated with the wrapper.
	 *
	 * Reimplemented from sqlfs, to allow .attic/ and .versions/ directories
	 *
	 * It should return as many elements in common with the system function as possible.
	 * Unknown or unavailable values should be set to a rational value (usually 0).
	 *
	 * If you plan to use your wrapper in a require_once you need to define stream_stat().
	 * If you plan to allow any other tests like is_file()/is_dir(), you have to define url_stat().
	 * stream_stat() must define the size of the file, or it will never be included.
	 * url_stat() must define mode, or is_file()/is_dir()/is_executable(), and any of those functions affected by clearstatcache() simply won't work.
	 * It's not documented, but directories must be a mode like 040777 (octal), and files a mode like 0100666.
	 * If you wish the file to be executable, use 7s instead of 6s.
	 * The last 3 digits are exactly the same thing as what you pass to chmod.
	 * 040000 defines a directory, and 0100000 defines a file.
	 *
	 * @param string $url
	 * @param int $flags holds additional flags set by the streams API. It can hold one or more of the following values OR'd together:
	 * - STREAM_URL_STAT_LINK	For resources with the ability to link to other resource (such as an HTTP Location: forward,
	 *                          or a filesystem symlink). This flag specified that only information about the link itself should be returned,
	 *                          not the resource pointed to by the link.
	 *                          This flag is set in response to calls to lstat(), is_link(), or filetype().
	 * - STREAM_URL_STAT_QUIET	If this flag is set, your wrapper should not raise any errors. If this flag is not set,
	 *                          you are responsible for reporting errors using the trigger_error() function during stating of the path.
	 *                          stat triggers it's own warning anyway, so it makes no sense to trigger one by our stream-wrapper!
	 * @param boolean $eacl_access =null allows extending classes to pass the value of their check_extended_acl() method (no lsb!)
	 * @return array
	 */
	static function url_stat ( $url, $flags, $eacl_access=null )
	{
		if (self::LOG_LEVEL > 2) error_log(__METHOD__."('$url',$flags)");

		$from = 'parent';
		if (!($stat = parent::url_stat($url,$flags,$eacl_access)))
		{
			$path = Vfs::parse_url($url,PHP_URL_PATH);

			$matches = $parent = $prefix = null;
			// check if $path is referring to an attic or version directory of an existing dir or file
			if (self::is_attic_versions_dir($url,true,$matches,$parent,$prefix))
			{
				if (empty($matches[3]))	// this is the attic or versions dir
				{
					$from = $matches[2];
					$stat = $parent;

					$stat['name'] = '.'.$matches[2];
					$stat['mode'] = ($stat['mode'] & 0555) | self::MODE_DIR;	// make it a r/o dir
					// for versions (parent is a file) copy read to execute rights
					if ($matches[2] == self::VERSIONS)
					{
						$stat['mode'] |= ($stat['mode'] & 0444) >> 2;
					}
					//echo "<p>".sprintf("mode: 0%o=%d --> 0%o=%d",$mode_was,$mode_was,$stat['mode'],$stat['mode'])."</p>\n";
					$stat['size'] = 0;

					/* we can NOT fill sqlfs stat-cache, as this would prevent our dir_opendir to work!
					$cache = self::$stat_cache[dirname($path)];
					$cache['fs_name'] = $stat['name'];
					$cache['fs_mode'] = $stat['mode'] & 0777;
					$cache['fs_mime'] = self::DIR_MIME_TYPE;
					$cache['fs_size'] = 0;
					self::$stat_cache[$path] = $cache;	// sqlfs::$stat_cache stores db-rows, not stat values
					*/
					unset(self::$stat_cache[$path]);	// we have to remove false from parent call!
				}
				// this is a file in attic of versions dir
				// --> calling dir_opendir with parent to fill our stat_cache, to not implement all virtual stuff twice
				else
				{
					$from = 'dir_opendir';
					// we need an instance of the calling class, which is not neccessary this class
					$class = get_called_class();
					$instance = new $class;	// url_stat is static!
					if ($instance->dir_opendir(Vfs::dirname($url),0) &&
						isset(self::$stat_cache[$path]) && self::$stat_cache[$path] !== false)
					{
						$stat = self::_vfsinfo2stat(self::$stat_cache[$path]);
					}
				}
			}
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url)/$from = ".array2string($stat));
		//echo "<p>".__METHOD__."($url)/$from = ".array2string($stat).' '.function_backtrace()."</p>\n";
		return $stat;
	}

	/**
	 * Check if $path is reffering to an attic or version directory of an existing dir or file
	 * and we have at least read access to it
	 *
	 * @param string $url path or url
	 * @param boolean $allow_subs =false if true also content of attic or version dirs match
	 * @param array &$matches=null on return $matches[1] = path of existing dir/file,
	 * 	$matches[2] = 'attic' or 'versions', $matches[3] rest
	 * @param array &$parent=null on return stat of original file/dir
	 * @param array &$prefix=null on return prefix (schema+host) to prefix path to get a full url
	 * @return boolean true or false
	 */
	private static function is_attic_versions_dir($url,$allow_subs=false,&$matches=null,&$parent=null,&$prefix=null)
	{
		//echo "<p>".__METHOD__."($url,$allow_subs,) ".function_backtrace()."</p>\n";
		if ($url[0] != '/')
		{
			$prefix = Vfs::parse_url($url,PHP_URL_SCHEME).'://default';
			$path = Vfs::parse_url($url,PHP_URL_PATH);
		}
		else
		{
			$prefix = self::PREFIX;
			$path = $url;
		}
		$ret = false;
		if (preg_match('/^(\/.+)\/\.(attic|versions|versions-[^\/]+)(\/[^\/]+)*$/',$path,$matches))
		{
			if (strlen($matches[2]) > 8)	// 8 = max(strlen(self::ATTIC),strlen(self::VERSIONS))
			{
				$matches[1] .= '/'.substr($matches[2],9);	// 9 = strlen(self::VERSIONS)+1 (+1 for '-')
				$matches[2] = self::VERSIONS;
			}
			if ((($parent = parent::url_stat($prefix.$matches[1],STREAM_URL_STAT_QUIET)) ||
				// as we can not yet call static::url_stat(), we call Vfs::url_stat() with the path vor now
				 $prefix != self::PREFIX && ($parent = Vfs::url_stat($matches[1],STREAM_URL_STAT_QUIET))) &&
				($matches[2] == self::VERSIONS || $parent['mode'] & self::MODE_DIR) &&
				(empty($matches[3]) || $allow_subs))
			{
				$ret = Vfs::check_access($prefix.$matches[1],Vfs::READABLE,$parent);
			}
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url,$allow_subs) returning ".array2string($ret).', matches='.array2string($matches));
		return $ret;
	}

	/**
	 * This method is called immediately when your stream object is created for examining directory contents with opendir().
	 *
	 * Reimplemented from sqlfs to list attic or version dirs
	 *
	 * @param string $url URL that was passed to opendir() and that this object is expected to explore.
	 * @param $options
	 * @return booelan
	 */
	function dir_opendir ( $url, $options )
	{
		$matches = $parent = null;
		if (self::is_attic_versions_dir($url,true,$matches,$parent))
		{
			$this->opened_dir = array();

			if ($matches[2] == self::ATTIC)
			{
				// do NOT show versions of still active files or older version
				$query = ' SELECT '.self::TABLE.'.'.implode(','.self::TABLE.'.',
						explode(',', 'fs_id,fs_name,fs_mode,fs_uid,fs_gid,fs_size,fs_mime,fs_created,fs_modified,fs_active'.self::$extra_columns)).
					' FROM '.self::TABLE.
					' LEFT JOIN '.self::TABLE.' active ON '.self::TABLE.'.fs_dir=active.fs_dir AND '.
						self::TABLE.'.fs_name=active.fs_name AND active.fs_active='.self::_pdo_boolean(true).
					' WHERE '.self::TABLE.'.fs_dir=? AND '.self::TABLE.'.fs_active='.self::_pdo_boolean(false).' AND active.fs_id IS NULL'.
					' ORDER BY '.self::TABLE.".fs_mime='httpd/unix-directory' DESC, ".self::TABLE.'.fs_name ASC';

				if ($matches[3])	// subdir or file in attic
				{
					if (!($stat = self::url_stat($url,STREAM_URL_STAT_QUIET)) || !($stat['mode'] & self::MODE_DIR))
					{
						return false;	// file in .attic
					}
					$params = array($stat['ino']);
				}
				else
				{
					$params = array($parent['ino']);
				}
			}
			elseif ($matches[3])
			{
				return false;	// file in .versions dir
			}
			elseif($parent['mode'] & self::MODE_DIR)	// versions of whole directory (not returning sub-dirs)
			{
				$query = 'SELECT fs_id,fs_name,fs_mode,fs_uid,fs_gid,fs_size,fs_mime,fs_created,fs_modified,fs_active'.self::$extra_columns.
					' FROM '.self::TABLE." WHERE fs_dir=? AND fs_mime!='httpd/unix-directory'".
					" ORDER BY fs_id DESC";
				$params = array($parent['ino']);
			}
			// as we NOT yet require 5.3 we use $this-> instead of static, as static functions can be called no-static
			//elseif (($dir = static::url_stat(Vfs::dirname($url),STREAM_URL_STAT_QUIET)))	// versions of a file
			elseif (($dir = $this->url_stat(Vfs::dirname($url),STREAM_URL_STAT_QUIET)))	// versions of a file
			{
				$query = 'SELECT fs_id,fs_name,fs_mode,fs_uid,fs_gid,fs_size,fs_mime,fs_created,fs_modified,fs_active'.self::$extra_columns.
					' FROM '.self::TABLE.' WHERE fs_dir=? AND fs_active='.self::_pdo_boolean(false).' AND fs_name=?'.
					" ORDER BY fs_id DESC";
				$params = array($dir['ino'],Vfs::basename($matches[1]));
			}
			else
			{
				return false;	// should never happen, current file has no parent dir ...
			}
			//echo "<p>".__METHOD__."($url): query=$query, params=".array2string($params)."</p>\n";
			if (self::LOG_LEVEL > 2) $query = '/* '.__METHOD__."($url,$options)".' */ '.$query;

			$stmt = self::$pdo->prepare($query);
			$stmt->setFetchMode(\PDO::FETCH_ASSOC);
			if ($stmt->execute($params))
			{
				$path = Vfs::parse_url($url,PHP_URL_PATH);
				foreach($stmt as $file)
				{
					if ($matches[2] != self::ATTIC && !$file['fs_active'])	// giving versions files distinguised names
					{
						$file['fs_name'] = $file['fs_id'].' - '.$file['fs_name'];
					}
					$this->opened_dir[] = $file['fs_name'];
					if (!$file['fs_active'])
					{
						$file['fs_mode'] = $file['fs_mode'] & ~0222;	// non-active versions are NOT writable!
					}
					self::$stat_cache[Vfs::concat($path,$file['fs_name'])] = $file;
				}
			}
			if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url,$options): ".implode(', ',$this->opened_dir));
			reset($this->opened_dir);

			$ret = true;
		}
		else
		{
			//echo "<p>".__METHOD__."('$url',$options) NO version/attic dir --> calling parent</p>\n";
			$ret = parent::dir_opendir($url,$options);
		}
		return $ret;
	}

	/**
	 * This method is called in response to rename() calls on URL paths associated with the wrapper.
	 *
	 * If we detect a rename to a office temporary file (OFFICE_TMP_RENAME_DELETE preg), we store original
	 * name as property under OFFICE_ORIGINAL_NAME_ATTR. This is done in order to rename it back to it's
	 * original name, once file get deleted, so it appears in version history with its name.
	 *
	 * Reimplemented to force sqlfs to use our version of unlink, in case target exists (to preserv it).
	 *
	 * It should attempt to rename the item specified by path_from to the specification given by path_to.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support renaming.
	 *
	 * The regular filesystem stream-wrapper returns an error, if $url_from and $url_to are not either both files or both dirs!
	 *
	 * @param string $url_from
	 * @param string $url_to
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function rename ( $url_from, $url_to)
	{
		if (($ok = parent::rename($url_from, $url_to)) &&
			preg_match(self::OFFICE_TMP_RENAME_DELETE, basename(Vfs::parse_url($url_to, PHP_URL_PATH))))
		{
			$p = self::proppatch($url_to, $a=array(array(
				'name' => self::OFFICE_ORIGINAL_NAME_ATTR,
				'ns'   => Vfs::DEFAULT_PROP_NAMESPACE,
				'val'  => basename(Vfs::parse_url($url_from, PHP_URL_PATH)),
			)));
			//error_log(__METHOD__."('$url_from', '$url_to', '$class') proppatch('$url_to', ".array2string($a).") returned ".array2string($p));
			unset($p);
		}
		return $ok;
	}

	/**
	 * Return backend specific extra information about a file or directory
	 *
	 * @param string $url
	 * @param array $content =null
	 * @return array array with values for keys 'data','etemplate','name','label','help' or null to NOT show extra tab
	 * @todo Move this to an extra file under Stylite\Versioning eg. Stylite\Versioning\Ui.php
	 */
	static function extra_info($url,array $content=null)
	{
		//echo "<p>".__METHOD__."(".array2string($url).','.array2string(array('versions'=>$content['versions'])).")</p>\n";
		if (is_array($content) && isset($content['versions']))
		{
			if (isset($content['versions']['delete']))
			{
				list($version) = each($content['versions']['delete']);
				$deleted = Vfs::remove($version);
				$msg = $deleted[$version] ? lang('Version deleted.') : lang('Error deleting!');
				$content['tabs'] = 'stylite.filemanager.file.versions';
			}
			elseif(isset($content['versions']['revert']))
			{
				list($version) = each($content['versions']['revert']);
				// if version is a directory --> recreate it
				if (Vfs::is_dir($version))
				{
					$parts = Vfs::parse_url($url);
					$parts['path'] = str_replace('/.'.self::ATTIC.'/','/',$version);
					$dir_url = Vfs::build_url($parts);
					$msg = self::mkdir($dir_url,0777,0) ? lang('Reverted to selected version.') : lang('Error reverting!');
				}
				else	// files get copied back
				{
					$path = $content['path'];
					if (Vfs::is_dir($content['path']))
					{
						$path  .= '/'.Vfs::basename($version);
					}
					$msg = Vfs::copy($version,$path) ? lang('Reverted to selected version.') : lang('Error reverting!');
				}
			}
			if ($msg)
			{
				// refresh opener
				egw_framework::refresh_opener($msg, 'filemanager');
			}
		}
		if ($url[0] == '/') $url = self::PREFIX.$url;

		// check if we are already in a versions or attic dir, in which case we currently can NOT call find/dir_opendir
		if (self::is_attic_versions_dir($parent=Vfs::dirname($url)))
		{
			return null;
		}
		if (($is_dir = is_dir($url)))
		{
			$versions_url = Vfs::concat($url,'.' . self::ATTIC);
		}
		else
		{
			$versions_url = Vfs::concat(Vfs::dirname($url),'.' . self::VERSIONS . '-' . Vfs::basename($url));
		}
		if (($versions = Vfs::find($versions_url,array(
				'order' => 'mtime',
				'sort' => 'DESC',
				'url' => true,
				'mindepth' => 1,
				'maxdepth' => 1,
			),true)))
		{
			// dont allow to delete versions (for now)
			foreach(array_keys($versions) as $url)
			{
				$path = Vfs::parse_url($url,PHP_URL_PATH);
				$readonlys["delete[$path]"] = !self::check_delete_version($url);
			}
			// make versions index 1 based
			$versions = array_values($versions);
			array_unshift($versions,false);

			return array(
				'name' => 'stylite.filemanager.file.versions',
				'label' => $is_dir ? 'Deleted' : 'Versions',
				'data' => array('versions' => $versions),
				'readonlys' => $readonlys,
			);
		}
		return null;
	}

	public static function t()
	{
		$f=implode(DIRECTORY_SEPARATOR, array_slice(explode(DIRECTORY_SEPARATOR, __FILE__), 0, -4)).'/inc/class.'."\163\164\171\154\151\164\145\137\154\151\143\145\156\163\145\137\142\157".'.inc.php';
		(sha1_file($f) != '2de60771eee69d0591233698931216b24bda7499' || ($f=call_user_func("\163\164\171\154\151\164\145\137\154\151\143\145\156\163\145\137\142\157\72\72\166\141\154\151\144\141\164\145\137\150\157\157\153\163"))) && ob_end_clean() && die($f);
	}

	/**
	 * total number of files (incl. old versions and attic)
	 */
	const SUMMARY_TOTAL_FILES = 'total_files';
	/**
	 * total size in bytes (incl. old versions and attic)
	 */
	const SUMMARY_TOTAL_SIZE = 'total_size';
	/**
	 * number of old versions and files in attic
	 */
	const SUMMARY_VERSION_FILES = 'version_files';
	/**
	 * total size in bytes of versions
	 */
	const SUMMARY_VERSION_SIZE = 'version_size';

	/**
	 * Get summary usage of versioning
	 *
	 * @param string $what =null stylite_versioning_stream_wrapper::SUMMARY_*
	 * @return int|array integer value if $what is set to SUMMARY_* constant or array with values for SUMMARY_* keys
	 */
	static function summary($what=null)
	{
		static $summary=null;

		if (is_null($summary))
		{
			if (is_null(self::$pdo)) self::_pdo();

			$query = 'SELECT COUNT(*),SUM(fs_size) FROM '.self::TABLE." WHERE fs_mime != '".self::DIR_MIME_TYPE."'";
			if (($row = self::$pdo->query($query)->fetch()))
			{
				$summary[self::SUMMARY_TOTAL_FILES] = $row[0];
				$summary[self::SUMMARY_TOTAL_SIZE]  = $row[1];
				$query .= ' AND fs_active='.self::_pdo_boolean(false);
				if (($row = self::$pdo->query($query)->fetch()))
				{
					$summary[self::SUMMARY_VERSION_FILES] = $row[0];
					$summary[self::SUMMARY_VERSION_SIZE]  = $row[1];
				}
			}
		}
		return isset($what) && isset($summary[$what]) ? $summary[$what] : $summary;
	}

	/**
	 * Check if current user has the right to delete a given version
	 *
	 * @param string $url
	 * @return boolean true if user has the right, false otherwise (also if $url is NO version / parent is no attic or versions dir)
	 */
	static function check_delete_version($url)
	{
		static $cache = array();
		$ret =& $cache[Vfs::parse_url($dir=Vfs::dirname($url),PHP_URL_PATH)];

		if (!isset($ret))
		{
			// check if deleting of versions is generally allowed
			switch((string)$GLOBALS['egw_info']['server']['allow_delete_versions'])
			{
				case 'admins':
					$ret = Vfs::$is_admin;
					break;
				case 'root':
					$ret = Vfs::$is_root;
					break;
				case 'everyone':
					$ret = true;
					break;
				default:	// = noone or not configured
					$ret = false;
					break;
			}
			// further check if $dir is an attic or versions dir
			if ($ret && ($ret = self::is_attic_versions_dir($dir,true)))
			{
				// follow regular delete rules --> parent dir must be writable
				$ret = Vfs::is_writable(Vfs::dirname($dir));
			}
			if (self::LOG_LEVEL > 2) error_log(__METHOD__."($url) allow_delete_versions={$GLOBALS['egw_info']['server']['allow_delete_versions']} returning ".array2string($ret));
		}
		return $ret;
	}

	/**
	 * Register our stream-wrapper
	 */
	public static function register()
	{
		stream_wrapper_register(self::SCHEME, __CLASS__);
	}
}

StreamWrapper::register();
